/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-1999.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <dirent.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include "fnmatch.h"

#include "gdk/gdkkeysyms.h"
#include "gtkbutton.h"
#include "gtkentry.h"
#include "gtkfilesel.h"
#include "gtkhbox.h"
#include "gtkhbbox.h"
#include "gtklabel.h"
#include "gtklist.h"
#include "gtklistitem.h"
#include "gtkmain.h"
#include "gtkscrolledwindow.h"
#include "gtksignal.h"
#include "gtkvbox.h"
#include "gtkmenu.h"
#include "gtkmenuitem.h"
#include "gtkoptionmenu.h"
#include "gtkclist.h"
#include "gtkdialog.h"
#include "gtkintl.h"
#include "gtkpixmap.h"

static char * home_xpm[] = {
"70 45 474 2",
"  	c None",
". 	c #000000",
"+ 	c #262623",
"@ 	c #242521",
"# 	c #22231F",
"$ 	c #21221E",
"% 	c #1F201C",
"& 	c #1D1E1A",
"* 	c #1C1D1A",
"= 	c #1B1C18",
"- 	c #191A16",
"; 	c #181915",
"> 	c #171814",
", 	c #D8D9D3",
"' 	c #F5F5F0",
") 	c #F1F2EC",
"! 	c #7B7C75",
"~ 	c #EBEBE5",
"{ 	c #F1F2EB",
"] 	c #E5E6D8",
"^ 	c #EEEFE6",
"/ 	c #E7E7E1",
"( 	c #1A1B17",
"_ 	c #ECECE2",
": 	c #DEDFCE",
"< 	c #ECEDE3",
"[ 	c #71716C",
"} 	c #E0E1D1",
"| 	c #D4D5CF",
"1 	c #2F302B",
"2 	c #E6E7DA",
"3 	c #F4F4EF",
"4 	c #D5D6C8",
"5 	c #20211C",
"6 	c #E3E4D5",
"7 	c #E7E8DB",
"8 	c #2B2C27",
"9 	c #DDDECD",
"0 	c #D9DACA",
"a 	c #D4D5C5",
"b 	c #D3D4C5",
"c 	c #D3D4C4",
"d 	c #D2D3C3",
"e 	c #D1D2C2",
"f 	c #DBDCCB",
"g 	c #C8C9BA",
"h 	c #9B9C90",
"i 	c #8B8C86",
"j 	c #8C8C88",
"k 	c #8D8D88",
"l 	c #8D8E88",
"m 	c #8E8E89",
"n 	c #8E8E8A",
"o 	c #8E8E8B",
"p 	c #8E8F8B",
"q 	c #8F908B",
"r 	c #90908C",
"s 	c #90918C",
"t 	c #91918C",
"u 	c #91918D",
"v 	c #91928E",
"w 	c #92928E",
"x 	c #93938F",
"y 	c #949490",
"z 	c #949491",
"A 	c #949591",
"B 	c #939490",
"C 	c #939390",
"D 	c #92928F",
"E 	c #91918E",
"F 	c #82827A",
"G 	c #C3C4B5",
"H 	c #DADBCA",
"I 	c #B6B7A9",
"J 	c #9C9C99",
"K 	c #FFFFFF",
"L 	c #FDFDFD",
"M 	c #FBFBFB",
"N 	c #F9F9F9",
"O 	c #F7F7F7",
"P 	c #F6F6F6",
"Q 	c #F4F4F4",
"R 	c #F2F2F2",
"S 	c #F0F0F0",
"T 	c #EEEEEE",
"U 	c #ECECEC",
"V 	c #C7C7C7",
"W 	c #9B9C91",
"X 	c #B0B1A4",
"Y 	c #BBBBBA",
"Z 	c #FEFEFE",
"` 	c #FCFCFC",
" .	c #FAFAFA",
"..	c #F8F8F8",
"+.	c #F3F3F3",
"@.	c #F1F1F1",
"#.	c #EFEFEF",
"$.	c #EDEDED",
"%.	c #EBEBEB",
"&.	c #E3E3E3",
"*.	c #88887F",
"=.	c #E9E9E9",
"-.	c #E1E1E1",
";.	c #F5F5F5",
">.	c #E8E8E8",
",.	c #E0E0E0",
"'.	c #E6E6E6",
").	c #DEDEDE",
"!.	c #EAEAEA",
"~.	c #E4E4E4",
"{.	c #DCDCDC",
"].	c #87877E",
"^.	c #DCDDCC",
"/.	c #2B2C26",
"(.	c #E2E2E2",
"_.	c #DADADA",
":.	c #85867E",
"<.	c #D8D9C9",
"[.	c #D6D6D0",
"}.	c #D7D7CD",
"|.	c #C6C7B8",
"1.	c #A1A296",
"2.	c #ABABA9",
"3.	c #E8E8E7",
"4.	c #E1E1E0",
"5.	c #E0E0DF",
"6.	c #D8D9D8",
"7.	c #D7D7D6",
"8.	c #D5D5D5",
"9.	c #D3D3D3",
"0.	c #D2D2D2",
"a.	c #D0D0D0",
"b.	c #CFCFCE",
"c.	c #CDCDCC",
"d.	c #C6C6C5",
"e.	c #7A7A73",
"f.	c #C4C5B6",
"g.	c #282924",
"h.	c #2D2E28",
"i.	c #2D2E27",
"j.	c #2C2E27",
"k.	c #2C2D27",
"l.	c #2B2D26",
"m.	c #242620",
"n.	c #5D5F52",
"o.	c #C5C7B4",
"p.	c #C4C7B4",
"q.	c #C4C6B4",
"r.	c #C3C5B3",
"s.	c #C3C5B2",
"t.	c #C2C5B2",
"u.	c #C2C4B2",
"v.	c #C1C4B1",
"w.	c #C0C3B0",
"x.	c #C0C2B0",
"y.	c #BFC2AF",
"z.	c #BEC1AF",
"A.	c #BEC0AE",
"B.	c #BDC0AE",
"C.	c #BDBFAE",
"D.	c #BDBFAD",
"E.	c #BCBEAD",
"F.	c #BBBEAC",
"G.	c #BBBDAC",
"H.	c #BABDAB",
"I.	c #B9BCAB",
"J.	c #B9BBAA",
"K.	c #B8BBA9",
"L.	c #B7BAA9",
"M.	c #B7B9A9",
"N.	c #B6B9A8",
"O.	c #B6B9A7",
"P.	c #B5B8A7",
"Q.	c #B5B7A7",
"R.	c #B4B7A6",
"S.	c #B3B6A6",
"T.	c #8F937C",
"U.	c #282922",
"V.	c #44463B",
"W.	c #BBBEAD",
"X.	c #BABCAB",
"Y.	c #B8BBAA",
"Z.	c #B7BAA8",
"`.	c #B6B8A7",
" +	c #B3B6A5",
".+	c #B2B5A4",
"++	c #B1B4A2",
"@+	c #B0B3A2",
"#+	c #AFB2A1",
"$+	c #ADB09E",
"%+	c #939781",
"&+	c #81856E",
"*+	c #191B17",
"=+	c #1F211B",
"-+	c #C1C3AF",
";+	c #C4C7B3",
">+	c #ADB095",
",+	c #AAAE92",
"'+	c #AAAD91",
")+	c #A9AD91",
"!+	c #A9AD90",
"~+	c #A9AC90",
"{+	c #A7AB8F",
"]+	c #A6AA8E",
"^+	c #A5A98D",
"/+	c #A5A98C",
"(+	c #A4A88B",
"_+	c #A2A68A",
":+	c #A1A589",
"<+	c #A0A487",
"[+	c #9FA387",
"}+	c #9DA186",
"|+	c #9DA184",
"1+	c #9CA084",
"2+	c #9B9F83",
"3+	c #9A9E82",
"4+	c #989C81",
"5+	c #979B80",
"6+	c #979B7F",
"7+	c #95997E",
"8+	c #94987E",
"9+	c #93977C",
"0+	c #92967B",
"a+	c #91957B",
"b+	c #90947A",
"c+	c #909479",
"d+	c #8F9378",
"e+	c #8E9278",
"f+	c #8D9177",
"g+	c #8C9076",
"h+	c #8B8F76",
"i+	c #8A8E75",
"j+	c #8A8E74",
"k+	c #898D74",
"l+	c #888C73",
"m+	c #878B72",
"n+	c #868A71",
"o+	c #858971",
"p+	c #656A59",
"q+	c #171714",
"r+	c #A9AB98",
"s+	c #C3C5B1",
"t+	c #A4A88A",
"u+	c #A3A789",
"v+	c #A2A688",
"w+	c #A1A588",
"x+	c #A1A587",
"y+	c #A0A486",
"z+	c #9FA386",
"A+	c #9EA285",
"B+	c #9CA083",
"C+	c #9B9F82",
"D+	c #999D81",
"E+	c #989C80",
"F+	c #969A7E",
"G+	c #94987D",
"H+	c #92967C",
"I+	c #91957A",
"J+	c #8F9379",
"K+	c #8E9277",
"L+	c #8B8F75",
"M+	c #898D73",
"N+	c #848870",
"O+	c #84886F",
"P+	c #4D5348",
"Q+	c #161713",
"R+	c #878978",
"S+	c #9A9E81",
"T+	c #95997D",
"U+	c #94987C",
"V+	c #92967A",
"W+	c #8D9176",
"X+	c #888C72",
"Y+	c #878B71",
"Z+	c #858970",
"`+	c #83876E",
" @	c #80856C",
".@	c #383F37",
"+@	c #161613",
"@@	c #606355",
"#@	c #A6AA8D",
"$@	c #A2A689",
"%@	c #9CA085",
"&@	c #9A9E84",
"*@	c #83876F",
"=@	c #82866E",
"-@	c #81856D",
";@	c #80846D",
">@	c #727966",
",@	c #282C28",
"'@	c #36382F",
")@	c #A7AB94",
"!@	c #C9CBBD",
"~@	c #C1C3B3",
"{@	c #A4A792",
"]@	c #93977D",
"^@	c #7F836C",
"/@	c #7F836B",
"(@	c #616A5D",
"_@	c #1A1A16",
":@	c #BCBEA9",
"<@	c #A9AD92",
"[@	c #969A7F",
"}@	c #A1A58C",
"|@	c #C8CABC",
"1@	c #C5C7B8",
"2@	c #989C82",
"3@	c #ACAF9C",
"4@	c #C0C3B4",
"5@	c #C6C8BA",
"6@	c #C5C7B9",
"7@	c #B2B5A3",
"8@	c #9A9D86",
"9@	c #7E826A",
"0@	c #7B8069",
"a@	c #4C574F",
"b@	c #151612",
"c@	c #9C9F8D",
"d@	c #9A9E86",
"e@	c #C7C9BB",
"f@	c #C6C9BB",
"g@	c #C5C8BA",
"h@	c #C5C7BA",
"i@	c #C4C6B9",
"j@	c #C3C5B8",
"k@	c #B7B9AA",
"l@	c #9DA18C",
"m@	c #7E826B",
"n@	c #7C8069",
"o@	c #6F7664",
"p@	c #424B44",
"q@	c #121310",
"r@	c #797B6C",
"s@	c #AAAD94",
"t@	c #95997F",
"u@	c #ABAE99",
"v@	c #C1C3B4",
"w@	c #C2C4B7",
"x@	c #BCBEAF",
"y@	c #A3A693",
"z@	c #8A8E76",
"A@	c #82866D",
"B@	c #81856C",
"C@	c #80846C",
"D@	c #7F836A",
"E@	c #626C5E",
"F@	c #363E37",
"G@	c #505247",
"H@	c #AAAE96",
"I@	c #999D84",
"J@	c #B0B3A0",
"K@	c #C4C6B8",
"L@	c #C3C5B9",
"M@	c #C2C4B8",
"N@	c #C1C3B7",
"O@	c #C1C3B6",
"P@	c #C0C2B6",
"Q@	c #BEC0B3",
"R@	c #B8BAAC",
"S@	c #A1A492",
"T@	c #878B74",
"U@	c #7D816A",
"V@	c #7A7F68",
"W@	c #57635A",
"X@	c #272C27",
"Y@	c #0D0E0C",
"Z@	c #282923",
"`@	c #ACAF97",
" #	c #C0C2B5",
".#	c #BFC1B5",
"+#	c #AEB1A1",
"@#	c #536058",
"##	c #ADAF9D",
"$#	c #ADB09A",
"%#	c #B3B5A4",
"&#	c #BEC0B4",
"*#	c #A5A898",
"=#	c #4A564F",
"-#	c #8C8E7E",
";#	c #AEB19B",
">#	c #AAAD9A",
",#	c #90947C",
"'#	c #868A72",
")#	c #90947E",
"!#	c #9DA08E",
"~#	c #131411",
"{#	c #0E0F0C",
"]#	c #696B5D",
"^#	c #AFB29D",
"/#	c #93977B",
"(#	c #8A8E73",
"_#	c #A1A490",
":#	c #949781",
"<#	c #85896F",
"[#	c #8E917A",
"}#	c #BFC1B4",
"|#	c #BDBFB2",
"1#	c #939683",
"2#	c #6E7563",
"3#	c #566259",
"4#	c #10100D",
"5#	c #080807",
"6#	c #414239",
"7#	c #B0B39F",
"8#	c #979B85",
"9#	c #979A85",
"0#	c #898D77",
"a#	c #BDBFB3",
"b#	c #BCBEB2",
"c#	c #8B8F7B",
"d#	c #626B5E",
"e#	c #0B0B09",
"f#	c #ADB09C",
"g#	c #8D907A",
"h#	c #9B9E8A",
"i#	c #878A75",
"j#	c #848772",
"k#	c #7A7E68",
"l#	c #191C18",
"m#	c #040404",
"n#	c #141511",
"o#	c #969A87",
"p#	c #838770",
"q#	c #A6A998",
"r#	c #A6A897",
"s#	c #919581",
"t#	c #7D806A",
"u#	c #4A554E",
"v#	c #11120F",
"w#	c #797B6D",
"x#	c #616B5E",
"y#	c #595B4E",
"z#	c #787D67",
"A#	c #767B67",
"B#	c #707664",
"C#	c #56635A",
"D#	c #10110E",
"E#	c #31332B",
"F#	c #8C9078",
"G#	c #737965",
"H#	c #717865",
"I#	c #6D7463",
"J#	c #6B7262",
"K#	c #687161",
"L#	c #666F60",
"M#	c #646D5F",
"N#	c #5F6A5D",
"O#	c #5D685C",
"P#	c #5B665C",
"Q#	c #59645B",
"R#	c #0D0E0B",
"S#	c #686B58",
"T#	c #7B7F68",
"U#	c #787C66",
"V#	c #757A65",
"W#	c #727763",
"X#	c #707662",
"Y#	c #6C7361",
"Z#	c #6A715F",
"`#	c #676E5E",
" $	c #656C5D",
".$	c #62695B",
"+$	c #60685A",
"@$	c #5D6559",
"#$	c #5A6458",
"$$	c #586257",
"%$	c #566156",
"&$	c #546056",
"*$	c #525E55",
"=$	c #505D54",
"-$	c #505C54",
";$	c #4D5851",
">$	c #181A16",
",$	c #161714",
"                                . + @ # $ % & * = - ; > >                                                                                   ",
"                              > , ' ' ' ' ' ' ' ' ' ' ) ! >                                                                                 ",
"                              > ~ { ] ] ] ] ] ] ] ] ] ^ / (                                                                                 ",
"                              > ~ _ : : : : : : : : : : < [                                                                                 ",
"                              > ~ _ : : : : : : : : : : } | 1 > > > > > > > > > > > > > > > > > > > > > > > > >                             ",
"                              > ~ _ : : : : : : : : : : : 2 3 ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' ' 3 4 5                           ",
"                              > ~ _ : : : : : : : : : : : : 6 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 6 : 8                           ",
"                              > ~ _ : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : 8                           ",
"                              > ~ _ 9 0 a b b b b b b b b b b c c c c c c c c c c d e e e e e e e e e e d 0 : : 8                           ",
"                              > ~ _ f g h i j k l m n n o p q r s t u v w x x y y z A z z y B B C C D E F G 9 : 8                           ",
"                              > ~ _ H I J K K K K K K K K K K K K K K K K K K K K K L M N O P Q R S T U V W f : 8                           ",
"                              > ~ _ H X Y K K K K K K K K K K K K K K K K K K K K Z `  ...P Q +.@.#.$.%.&.*.f : 8                           ",
"                              > ~ _ H X Y K K K K K K K K K K K K K K K K K K K Z `  ...P Q +.@.#.$.%.=.-.*.f : 8                           ",
"                              > ~ _ H X Y K K K K K K K K K K K K K K K K K K Z `  ...O ;.+.@.#.$.%.=.>.,.*.f : 8                           ",
"                              > ~ _ H X Y K K K K K K K K K K K K K K K K K Z `  ...O ;.+.@.#.$.%.=.>.'.).*.f : 8                           ",
"                              > ~ _ H X Y K K K K K K K K K K K K K K K K K L M N O ;.+.R S T U !.>.'.~.{.].H ^./.                          ",
"                              > ~ _ H X Y K K K K K K K K K K K K K K K K L M N O ;.+.R S T U !.>.'.~.(._.:.<.f /.                          ",
"                              > [.}.|.1.2.3.3.3.3.3.3.3.3.3.3.3.3.3.3.3.'.~.(.4.5.).{._.6.7.8.9.0.a.b.c.d.e.f.|.g.                          ",
"                    > h.h.h.h.h.h.h.h.h.h.h.h.h.h.h.h.h.h.h.h.h.h.h.i.i.i.j.j.j.j.k.k.k.k.k.k.k.k.l.l.l.l./././././././.m.>                 ",
"                    n.o.o.o.o.o.o.o.o.o.o.o.o.o.o.o.o.o.p.q.r.s.t.u.v.w.x.x.y.z.A.B.C.D.E.F.G.H.I.I.J.K.L.M.N.O.P.Q.R.S.T.U.                ",
"                    V.o.o.o.o.o.o.o.o.o.o.o.o.o.o.o.p.q.r.s.t.u.v.w.x.x.y.z.A.B.C.E.E.W.H.X.Y.L.Z.O.`.R. +.+.+++@+#+$+%+&+*+                ",
"                    =+-+o.;+>+,+,+'+'+)+!+!+~+~+{+]+^+/+(+_+:+<+[+}+|+1+2+3+4+5+6+7+8+9+0+a+b+c+d+e+f+g+h+i+j+k+l+m+n+o+p+>                 ",
"                    q+r+o.s+t+t+t+t+t+t+t+u+v+w+x+y+z+A+|+1+B+C+3+D+E+5+6+F+7+G+9+H+0+I+b+J+e+K+f+g+L+i+k+M+l+m+n+o+N+O+P+>                 ",
"                    Q+R+o.p.t+t+t+t+t+u+v+w+x+y+z+A+|+B+B+C+S+D+E+6+6+F+T+U+9+0+V+I+c+d+e+K+W+g+L+j+M+M+X+Y+n+Z+O+O+`+ @.@+@                ",
"                      @@o.o.#@t+t+u+$@w+x+y+z+A+|+1+2+C+3+D+3+%@&@F+T+G+9+0+a+b+c+d+e+f+g+h+i+j+k+l+m+n+o+Z+O+*@=@-@;@>@,@                  ",
"                      '@p.o.~+u+$@w+x+y+z+A+|+1+2+C+3+D+E+6+)@!@~@G+9+H+{@]@c+d+e+f+g+h+i+j+k+l+m+n+o+Z+O+*@=@-@;@^@/@(@*+                  ",
"                      _@:@p.<@w+x+y+z+A+|+1+2+C+3+D+E+6+[@F+}@|@1@2@3@4@5@6@7@8@g+h+i+j+k+l+m+n+o+Z+O+*@=@-@;@^@/@9@0@a@>                   ",
"                      b@c@s.<@z+A+|+1+B+C+3+D+E+5+6+F+7+G+9+d@e@f@g@g@h@6@i@i@j@k@l@i+l+m+n+o+N+O+*@=@-@;@^@/@m@n@n@o@p@b@                  ",
"                      q@r@u.s@|+B+B+C+S+D+E+6+6+F+T+U+9+t@u@v@5@g@h@6@i@i@j@j@w@w@w@x@y@z@O+O+`+A@B@C@/@D@9@n@n@n@n@E@F@q@                  ",
"                        G@w.H@2+C+3+D+E+6+[@F+T+G+9+I@J@K@5@g@h@i@i@L@j@j@M@w@N@O@P@P@Q@R@S@T@;@^@/@9@U@n@n@n@n@n@V@W@X@Y@                  ",
"                        Z@A.`@3+D+E+6+[@F+T+G+9+0+a+b+c+d+x@i@i@L@j@j@M@w@N@O@P@P@ #.#+#-@;@^@/@9@U@n@n@n@n@n@n@n@o@@#*+                    ",
"                        b@##$#E+6+[@F+T+G+9+0+a+b+c+d+e+f+%#L@j@j@M@w@N@O@P@P@ #.#.#&#*#^@/@9@U@n@n@n@n@n@n@n@n@n@E@=#Q+                    ",
"                        q@-#;#[@7+T+G+9+0+a+b+c+d+e+f+g+h+>#j@M@,#m+'#)#P@P@.#.#&#&#Q@!#m@n@n@n@n@n@n@n@n@n@n@n@V@W@p@~#                    ",
"                        {#]#^#G+U+/#V+I+c+d+d+K+W+g+L+j+(#_#w@O@:#Z+<#[#.#}#Q@Q@Q@|#|#1#n@n@n@n@n@n@n@n@n@n@n@n@2#3#F@4#                    ",
"                        5#6#7#0+a+b+c+J+e+f+g+h+L+j+k+l+m+8#P@P@9#*@=@0#&#a#a#|#b#b#b#c#n@n@n@n@n@n@n@n@n@n@n@n@d#3#X@e#                    ",
"                          & f#b+J+d+e+f+g+L+i+j+M+l+m+n+o+g# #.#h#-@C@i#a#|#b#b#b#b#b#j#n@n@n@n@n@n@n@n@n@n@n@k#W@@#l#m#                    ",
"                          n#o#e+e+f+g+h+i+k+k+l+m+n+o+N+O+p#q#r#s#^@/@&+!#c@c@c@c@c@c@t#n@n@n@n@n@n@n@n@n@n@n@2#3#u#Q+                      ",
"                          v#w#f+g+h+i+k+M+l+m+n+o+N+O+*@=@-@;@^@/@m@n@n@n@n@n@n@n@n@n@n@n@n@n@n@n@n@n@n@n@n@n@x#3#p@~#                      ",
"                          Y@y#h+j+M+l+X+Y+n+Z+O+O+`+A@B@C@/@D@9@n@n@n@n@n@n@n@n@n@n@n@n@n@n@n@n@n@n@n@V@z#A#B#C#3#F@D#                      ",
"                            E#F#l+m+n+o+N+O+*@=@-@;@^@/@9@U@n@n@n@n@V@z#A#G#H#o@I#J#K#L#M#d#N#O#P#Q#C#3#3#3#3#3#3#X@R#                      ",
"                            ; S#T#U#V#W#X#Y#Z#`# $.$+$@$#$$$%$&$*$=$-$-$-$-$-$-$-$-$-$-$-$-$-$-$-$-$-$-$-$-$-$-$;$>$                        ",
"                              Q+> > > > > > > > > > > > > > > > > > > > > > > > > > > > > > > > > > > > > > > > ,$                          ",
"                                                                                                                                            ",
"                                                                                                                                            ",
"                                                                                                                                            "};

static char * desktop_xpm[] = {
"69 45 434 2",
"  	c None",
". 	c #000000",
"+ 	c #B6BECA",
"@ 	c #A5AFBE",
"# 	c #737D8C",
"$ 	c #909CAF",
"% 	c #556784",
"& 	c #44526A",
"* 	c #4D4E41",
"= 	c #C8C9BC",
"- 	c #DBDCCC",
"; 	c #D3D4C0",
"> 	c #D2D3C0",
", 	c #D1D3BF",
"' 	c #D1D3BE",
") 	c #D1D2BE",
"! 	c #D0D1BD",
"~ 	c #CFD0BC",
"{ 	c #CED0BC",
"] 	c #CDD0BB",
"^ 	c #CDCFBA",
"/ 	c #CDCEB9",
"( 	c #CDCDB9",
"_ 	c #CCCDB9",
": 	c #71725E",
"< 	c #616252",
"[ 	c #97987F",
"} 	c #BCBE9F",
"| 	c #BBBD9E",
"1 	c #BBBD9D",
"2 	c #BABC9D",
"3 	c #B9BB9C",
"4 	c #B8BB9B",
"5 	c #B8BA9A",
"6 	c #B7B999",
"7 	c #B6B899",
"8 	c #B5B898",
"9 	c #B5B797",
"0 	c #B4B696",
"a 	c #B3B695",
"b 	c #B3B595",
"c 	c #B2B494",
"d 	c #B1B493",
"e 	c #B0B392",
"f 	c #B0B291",
"g 	c #AFB291",
"h 	c #6F725C",
"i 	c #B9BC9C",
"j 	c #B9BB9B",
"k 	c #B8BA9B",
"l 	c #B7B99A",
"m 	c #B6B898",
"n 	c #B4B797",
"o 	c #B2B594",
"p 	c #B1B393",
"q 	c #AFB190",
"r 	c #AEB08F",
"s 	c #ADB08F",
"t 	c #6E705B",
"u 	c #BBBCAC",
"v 	c #BABD9D",
"w 	c #B2B595",
"x 	c #AFB191",
"y 	c #AEB190",
"z 	c #ADAF8E",
"A 	c #ACAF8D",
"B 	c #96997C",
"C 	c #B6B999",
"D 	c #B0B292",
"E 	c #ACAF8E",
"F 	c #ACAE8D",
"G 	c #ABAE8C",
"H 	c #96987A",
"I 	c #7A7B67",
"J 	c #5B5C4C",
"K 	c #B7BA9A",
"L 	c #AAAD8C",
"M 	c #AAAD8B",
"N 	c #959779",
"O 	c #D3D3D3",
"P 	c #F6F6F6",
"Q 	c #FFFFFF",
"R 	c #F9F9F9",
"S 	c #DADADA",
"T 	c #585858",
"U 	c #2C2D25",
"V 	c #B3B596",
"W 	c #B4B796",
"X 	c #B3B696",
"Y 	c #B1B392",
"Z 	c #ADB08E",
"` 	c #ABAE8D",
" .	c #A9AC8A",
"..	c #A9AB8A",
"+.	c #949678",
"@.	c #C7C7C7",
"#.	c #D1D1D1",
"$.	c #D6D6D6",
"%.	c #B0B293",
"&.	c #AAAC8B",
"*.	c #A8AB89",
"=.	c #A7AA88",
"-.	c #939678",
";.	c #FEFEFE",
">.	c #FDFDFD",
",.	c #C0C0C0",
"'.	c #E1E1E1",
").	c #F0F0F0",
"!.	c #9B9B9B",
"~.	c #ADAF90",
"{.	c #ABAD8C",
"].	c #A8AB8A",
"^.	c #A6A987",
"/.	c #929576",
"(.	c #FCFCFB",
"_.	c #FBFBFB",
":.	c #AFAFAE",
"<.	c #E9E9E9",
"[.	c #DFDFDF",
"}.	c #8F8F8F",
"|.	c #ABAE8E",
"1.	c #AEB18F",
"2.	c #A9AC8B",
"3.	c #A7AA87",
"4.	c #A5A886",
"5.	c #909475",
"6.	c #FAFAF9",
"7.	c #F9F9F8",
"8.	c #A4A4A3",
"9.	c #F4F4F4",
"0.	c #CFCFCF",
"a.	c #A2A2A2",
"b.	c #34342B",
"c.	c #AAAD8D",
"d.	c #A5A986",
"e.	c #A5A885",
"f.	c #A4A785",
"g.	c #909274",
"h.	c #F8F8F7",
"i.	c #F8F7F6",
"j.	c #9E9E9E",
"k.	c #565747",
"l.	c #6F715C",
"m.	c #0D0D0B",
"n.	c #392E27",
"o.	c #8C8C70",
"p.	c #A5A787",
"q.	c #A6AA87",
"r.	c #A4A784",
"s.	c #A3A683",
"t.	c #8F9173",
"u.	c #F7F6F5",
"v.	c #F6F6F4",
"w.	c #F4F3F2",
"x.	c #DEDDDC",
"y.	c #D3D2D0",
"z.	c #B7B7B5",
"A.	c #9F9E9D",
"B.	c #706F6F",
"C.	c #65625A",
"D.	c #A2A486",
"E.	c #6F705B",
"F.	c #141411",
"G.	c #453F3F",
"H.	c #E7ACA3",
"I.	c #C25B4B",
"J.	c #2C1813",
"K.	c #818469",
"L.	c #A3A684",
"M.	c #A2A683",
"N.	c #A2A582",
"O.	c #8E9071",
"P.	c #F5F4F3",
"Q.	c #F2F2F0",
"R.	c #E4E4E2",
"S.	c #DAD9D7",
"T.	c #D8D8D6",
"U.	c #CDCCCA",
"V.	c #AFAEAC",
"W.	c #88847B",
"X.	c #11110C",
"Y.	c #444035",
"Z.	c #E8DAB0",
"`.	c #A09993",
" +	c #9D4E42",
".+	c #6E2316",
"++	c #200F0A",
"@+	c #6F725B",
"#+	c #A6A986",
"$+	c #A3A784",
"%+	c #A1A582",
"&+	c #A1A481",
"*+	c #8D9070",
"=+	c #F3F3F1",
"-+	c #EFEFED",
";+	c #EEEDEB",
">+	c #EDECEA",
",+	c #E9E8E6",
"'+	c #D5D4D3",
")+	c #BDBCBB",
"!+	c #110F0D",
"~+	c #2C2A23",
"{+	c #F0DFAB",
"]+	c #D0B466",
"^+	c #BA8C2C",
"/+	c #785F32",
"(+	c #341813",
"_+	c #191612",
":+	c #82856A",
"<+	c #A2A583",
"[+	c #A0A480",
"}+	c #9FA380",
"|+	c #8C8F70",
"1+	c #F6F5F4",
"2+	c #F5F5F3",
"3+	c #F1F1EF",
"4+	c #F1F0EE",
"5+	c #ECEBE9",
"6+	c #EAE9E7",
"7+	c #DAD6CD",
"8+	c #161511",
"9+	c #2D2C26",
"0+	c #F2E2B2",
"a+	c #D0B465",
"b+	c #B8892B",
"c+	c #A47114",
"d+	c #4F3406",
"e+	c #1B170D",
"f+	c #73755D",
"g+	c #9C9F7F",
"h+	c #A0A481",
"i+	c #A0A380",
"j+	c #9FA27F",
"k+	c #9EA27E",
"l+	c #8B8E6F",
"m+	c #BABC9C",
"n+	c #F3F2F1",
"o+	c #F0F0EE",
"p+	c #F0EFED",
"q+	c #EFEEEC",
"r+	c #C5C3BC",
"s+	c #15140E",
"t+	c #F2E3B3",
"u+	c #D0B567",
"v+	c #B98D2F",
"w+	c #A77216",
"x+	c #5A3D09",
"y+	c #1B1507",
"z+	c #5B5D4A",
"A+	c #9B9E7E",
"B+	c #9EA17E",
"C+	c #9DA17D",
"D+	c #898D6D",
"E+	c #3C382F",
"F+	c #2B281E",
"G+	c #F0DFA7",
"H+	c #D0B468",
"I+	c #BE9032",
"J+	c #B07C1B",
"K+	c #60410B",
"L+	c #181001",
"M+	c #3D3E31",
"N+	c #95987A",
"O+	c #9DA07D",
"P+	c #9CA07C",
"Q+	c #888C6C",
"R+	c #E7E6E4",
"S+	c #4E4A42",
"T+	c #AEA7A0",
"U+	c #CDB591",
"V+	c #C59E40",
"W+	c #B7821E",
"X+	c #62440E",
"Y+	c #1E1707",
"Z+	c #3A3834",
"`+	c #8C8E71",
" @	c #9FA37F",
".@	c #9C9F7B",
"+@	c #9B9F7B",
"@@	c #878B6B",
"#@	c #EDEDEB",
"$@	c #EBEAE8",
"%@	c #E9E8E5",
"&@	c #1F1F1F",
"*@	c #262420",
"=@	c #907E69",
"-@	c #9A7A55",
";@	c #7B6033",
">@	c #161007",
",@	c #444443",
"'@	c #B8B7B4",
")@	c #9B958A",
"!@	c #8E9073",
"~@	c #9A9E7A",
"{@	c #9A9D79",
"]@	c #878A6A",
"^@	c #E8E7E5",
"/@	c #E8E7E4",
"(@	c #E7E6E3",
"_@	c #1E1D1D",
":@	c #141414",
"<@	c #161613",
"[@	c #22211F",
"}@	c #4C4C4A",
"|@	c #B1B0AD",
"1@	c #D2D1CD",
"2@	c #D9D8D4",
"3@	c #9F998E",
"4@	c #8F9274",
"5@	c #9EA27F",
"6@	c #9B9E7B",
"7@	c #999D79",
"8@	c #999C78",
"9@	c #868969",
"0@	c #E6E5E2",
"a@	c #E5E4E1",
"b@	c #E6E4E1",
"c@	c #E5E4E0",
"d@	c #E0DEDB",
"e@	c #D0CFCB",
"f@	c #C2C2BE",
"g@	c #C3C3BF",
"h@	c #D8D7D3",
"i@	c #D8D6D2",
"j@	c #D9D7D3",
"k@	c #9F998D",
"l@	c #8E9173",
"m@	c #9DA07C",
"n@	c #9C9F7C",
"o@	c #999D78",
"p@	c #989C78",
"q@	c #979B77",
"r@	c #858868",
"s@	c #E4E3E0",
"t@	c #E3E2DF",
"u@	c #E3E2DE",
"v@	c #E4E2DF",
"w@	c #E4E3DF",
"x@	c #E2E1DD",
"y@	c #E0DFDB",
"z@	c #DBDAD6",
"A@	c #DAD8D4",
"B@	c #DCDAD6",
"C@	c #9E988D",
"D@	c #8D9071",
"E@	c #9B9E7A",
"F@	c #989C77",
"G@	c #979B76",
"H@	c #969A76",
"I@	c #848767",
"J@	c #EDEDED",
"K@	c #E1E0DD",
"L@	c #E1DFDC",
"M@	c #E1E0DC",
"N@	c #DFDEDA",
"O@	c #E0DEDA",
"P@	c #DDDBD7",
"Q@	c #D8D6D1",
"R@	c #D7D5D1",
"S@	c #9D978B",
"T@	c #969A75",
"U@	c #959974",
"V@	c #828666",
"W@	c #E1DFDB",
"X@	c #DEDDD9",
"Y@	c #DEDCD8",
"Z@	c #D7D6D1",
"`@	c #D5D3CE",
" #	c #969975",
".#	c #949873",
"+#	c #828565",
"@#	c #999891",
"##	c #A39E92",
"$#	c #A49E93",
"%#	c #A39D92",
"&#	c #A39D91",
"*#	c #A19C90",
"=#	c #A29C90",
"-#	c #A19B8F",
";#	c #A19B90",
">#	c #9D978C",
",#	c #9B968A",
"'#	c #676359",
")#	c #8B8D6F",
"!#	c #979A76",
"~#	c #949773",
"{#	c #CDCFB9",
"]#	c #6D6F59",
"^#	c #575A46",
"/#	c #8E9171",
"(#	c #989B77",
"_#	c #959975",
":#	c #D6D7C6",
"<#	c #A0A282",
"[#	c #95987B",
"}#	c #949779",
"|#	c #929577",
"1#	c #929476",
"2#	c #919476",
"3#	c #919376",
"4#	c #909275",
"5#	c #8F9273",
"6#	c #8E9172",
"7#	c #8D9072",
"8#	c #8C9071",
"9#	c #8A8D6F",
"0#	c #8A8C6E",
"a#	c #979A78",
"b#	c #76849C",
"c#	c #72725E",
"d#	c #7A7D5F",
"e#	c #9298A2",
"f#	c #70725D",
"g#	c #8B8D72",
"h#	c #8F9075",
"i#	c #979A7C",
"j#	c #96997B",
"k#	c #96987B",
"l#	c #919475",
"m#	c #909374",
"n#	c #8F9172",
"o#	c #8C9070",
"p#	c #8A8E6E",
"q#	c #8A8D6D",
"r#	c #898C6D",
"s#	c #878A69",
"t#	c #858869",
"u#	c #838766",
"v#	c #838666",
"w#	c #828665",
"x#	c #364155",
"y#	c #4B5564",
"z#	c #848C98",
"A#	c #2B3444",
"                                                                                                                                          ",
"                                                                                                                                          ",
"                                                                                                                                          ",
"                                                                                                                                          ",
"                                                                                                                                          ",
"                  . . . .                                                                                   . . . . .                     ",
"                . + @ @ # . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . + $ $ # .                   ",
"                . $ % & * = - ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; > , ' ) ! ! ! ~ { ] ^ / / ( _ : @ % & .                   ",
"                . $ & < [ } } } } } } } } } } } } } } } } } } } } } } } } } | 1 2 3 4 5 6 7 8 9 0 a b c d e f g h $ & .                   ",
"                . # < [ } } } } } } } } } } } } } } } } } } } } } } } } } | 2 i j k l 6 m 9 n 0 b o d p e g q r s t $ .                   ",
"                  . u } } } } } } } } } } } } } } } } } } } } } } } } | v 2 3 k 5 6 7 8 9 0 a w c d e f x y s z A B . .                   ",
"                  . ; } } } } } } } } } } } } } } } } } } } } } } } | 2 3 j k l C m 9 n 0 b o d p D g y r s E F G H .                     ",
"                  . ; } } } } } } } } } I . . . . . . . . . . . . J l 3 k K 6 m 8 9 0 a o c p e g q y s z A G L M N .                     ",
"                  . ; } } } } } } } } } . O P Q Q Q Q Q Q Q Q R S T U V 7 C m 9 W X b c d Y D g y r Z E ` G M  ...+..                     ",
"                  . ; } } } } } } } } } . P Q Q Q Q Q Q Q Q Q Q @.#.$.. %.n n 0 b o c p e g q y s E A G L &. .*.=.-..                     ",
"                  . ; } } } } } } } } } . Q Q Q Q Q Q Q ;.Q ;.>.,.'.).!.. ~.w b c d e D g y r Z E ` {.M  .].*.=.^./..                     ",
"                  . ; } } } } } } } } } . Q Q Q Q Q ;.>.(.>.(._.:.<.Q [.}.. |.e p e g q 1.s E F G M 2. .*.=.3.^.4.5..                     ",
"                  . ; } } } } } } } } } . Q Q Q ;.>.(._.6._.6.7.8.).Q 9.0.a.b.c.q g y s z A ` {.M  .*.*.=.^.d.e.f.g..                     ",
"                  . ; } } } } } } } } } . Q ;.>.(._.6.7.h.7.h.i.j.. . . . . . k.c.r s l.m.n.o.p. .*.=.q.^.4.f.r.s.t..                     ",
"                  . ; } } } } } } } } } . Q (._.6.7.h.i.u.i.u.v.w.x.y.z.A.B.C.. D.E.F.G.H.I.J.K.*.=.^.4.e.f.L.M.N.O..                     ",
"                  . ; } } } } } } } } | . Q 6.7.h.i.u.v.P.v.P.w.Q.R.S.T.U.V.W.. X.Y.Z.`. +.+++@+^.#+4.f.$+s.N.%+&+*+.                     ",
"                  . ; } } } } } } } | 2 . Q h.i.u.v.P.w.=+w.=+Q.-+;+>+,+'+)+!+~+{+]+^+/+(+_+:+<+4.e.r.L.M.N.&+[+}+|+.                     ",
"                  . ; } } } } } | | 2 3 . Q 1+2+w.w.=+Q.3+Q.3+4+;+5+6+7+8+9+0+a+b+c+d+e+f+g+r.L.f.$+s.N.%+h+i+j+k+l+.                     ",
"                  . ; } } } } | 2 m+3 k . Q w.n+Q.o+-+4+p+4+p+q+5+r+s+9+t+u+v+w+x+y+z+A+<+L.M.r.L.M.N.&+i+}+j+B+C+D+.                     ",
"                  . ; } } | 1 2 3 4 5 6 . Q w.n+Q.o+-+4+p+4+p+,+E+F+G+H+I+J+K+L+M+N+<+$+s.f.$+s.N.&+h+i+j+k+C+O+P+Q+.                     ",
"                  . ; } | 2 i j k l 6 m . Q o+-+-+q+;+>+>+>+R+S+T+U+V+W+X+Y+Z+. `+f.L.M.r.s.<+%+&+i+ @j+B+C+P+.@+@@@.                     ",
"                  . > v 2 3 k 5 6 7 8 9 . Q #@>+5+5+5+$@%@$@&@*@=@-@;@>@,@'@)@. !@s.f.L.M.N.&+h+i+j+k+C+O+P++@~@{@]@.                     ",
"                  . , 3 j k l C m 9 n 0 . Q $@,+,+^@/@/@(@_@:@:@<@[@}@|@1@2@3@. 4@r.s.<+%+&+i+ @5@B+C+P+.@6@~@7@8@9@.                     ",
"                  . ) k K 6 m 8 9 0 a o . Q (@0@0@0@a@b@c@d@e@f@g@e@h@2@i@j@k@. l@M.N.&+[+}+j+k+C+m@n@+@~@{@o@p@q@r@.                     ",
"                  . ! l C m 9 W X b c d . R s@t@u@v@w@x@u@x@y@z@z@z@A@B@j@i@C@. D@%+&+i+j+k+B+O+P++@E@~@7@8@F@G@H@I@.                     ",
"                  . ! m 8 n 0 b o c p e . J@K@L@M@y@x@y@M@y@M@N@O@P@B@j@Q@R@S@. |+i+}+j+B+C+P+n@+@~@{@o@p@q@G@T@U@V@.                     ",
"                  . ~ 9 0 X b c d e D g . #.W@O@N@X@y@X@O@X@O@Y@Y@B@B@i@Z@`@)@. l+j+k+B+O+P++@E@~@7@p@F@G@H@ #U@.#+#.                     ",
"                  . ] 0 b o d p e g q 1.. @###$#%#&#%#*#=#*#=#-#;#k@k@>#>#,#'#. )#B+C+P+.@+@~@7@o@F@q@!#T@U@.#.#~#+#.                     ",
"                  . {#b c d e f g y s z ]#. . . . . . . . . . . . . . . . . . ^#/#O+P++@E@{@7@p@(#G@T@_#U@.#~#~#~#+#.                     ",
"                . . :#d p e g y r s E F  .<#[#}#|#1#2#3#4#4@5#6#7#8#|+l+l+9#0#/#a#.@6@~@7@8@F@q@H@T@U@.#.#~#~#~#~#+#. .                   ",
"                . b#c#e f x y s z A ` L M  .*.*.=.^.4.e.f.L.M.N.&+h+i+j+k+C+O+n@+@~@{@7@p@(#G@T@_#U@.#~#~#~#~#~#~#d#e#.                   ",
"                . % b#f#y r Z E ` G M  . .*.=.^.#+4.f.$+s.N.%+&+i+ @k+B+O+P++@E@~@7@8@F@G@H@T@U@.#~#~#~#~#~#~#~#~#+ & .                   ",
"                . % % b#g#h#i#j#k#H N +.-.-.|#2#l#m#g.5#n#O.*+o#|+l+p#q#r#Q+@@]@s#9@t#r@I@u#v#w#+#+#+#+#+#+#+#+#@ % & .                   ",
"                . x#x#x#y#. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . z#x#x#A#.                   ",
"                . . . . . .                                                                                 . . . . . .                   ",
"                                                                                                                                          ",
"                                                                                                                                          ",
"                                                                                                                                          ",
"                                                                                                                                          ",
"                                                                                                                                          "};

static char * documents_xpm[] = {
"48 48 213 2",
"  	c None",
". 	c #000000",
"+ 	c #D3D3D3",
"@ 	c #F6F6F6",
"# 	c #FFFFFF",
"$ 	c #F9F9F9",
"% 	c #F9F9F8",
"& 	c #E7E7E7",
"* 	c #E4E4E4",
"= 	c #CDCDCD",
"- 	c #EFEFEE",
"; 	c #AFAFAE",
"> 	c #E9E9E9",
", 	c #B3B3B3",
"' 	c #797979",
") 	c #787878",
"! 	c #EEEEEE",
"~ 	c #A4A4A3",
"{ 	c #F0F0F0",
"] 	c #F4F4F4",
"^ 	c #ABABAB",
"/ 	c #FEFEFE",
"( 	c #FDFDFD",
"_ 	c #FCFCFB",
": 	c #FBFBFB",
"< 	c #F2F2F2",
"[ 	c #9E9E9E",
"} 	c #777777",
"| 	c #767676",
"1 	c #767675",
"2 	c #D4D4D3",
"3 	c #B9B7B6",
"4 	c #838381",
"5 	c #848381",
"6 	c #6A6763",
"7 	c #FAFAF9",
"8 	c #F8F8F7",
"9 	c #F8F7F6",
"0 	c #F3F2F1",
"a 	c #E0E0DF",
"b 	c #DCDAD8",
"c 	c #CACAC8",
"d 	c #B6B6B4",
"e 	c #72706C",
"f 	c #757575",
"g 	c #757574",
"h 	c #757474",
"i 	c #747473",
"j 	c #F4F3F2",
"k 	c #F1F0EF",
"l 	c #F0EFED",
"m 	c #E5E4E2",
"n 	c #C4C3C0",
"o 	c #817D74",
"p 	c #F7F6F5",
"q 	c #F6F6F4",
"r 	c #F5F4F3",
"s 	c #EBEAE8",
"t 	c #E1DFDC",
"u 	c #8B867D",
"v 	c #747373",
"w 	c #737372",
"x 	c #727271",
"y 	c #717170",
"z 	c #70706F",
"A 	c #6C6C6B",
"B 	c #6A6968",
"C 	c #A49E93",
"D 	c #F6F5F4",
"E 	c #F5F5F3",
"F 	c #F3F3F1",
"G 	c #F2F2F0",
"H 	c #F1F1EF",
"I 	c #F1F0EE",
"J 	c #EEEDEB",
"K 	c #ECEBE9",
"L 	c #E7E6E3",
"M 	c #E2E1DF",
"N 	c #E4E3E0",
"O 	c #737272",
"P 	c #727170",
"Q 	c #71706F",
"R 	c #6F6F6E",
"S 	c #6E6D6C",
"T 	c #6C6C6A",
"U 	c #6C6B6A",
"V 	c #6B6A69",
"W 	c #F0F0EE",
"X 	c #EFEFED",
"Y 	c #EFEEEC",
"Z 	c #E8E7E5",
"` 	c #E5E4E1",
" .	c #E2E1DE",
"..	c #98A8B5",
"+.	c #98A8B4",
"@.	c #98A7B4",
"#.	c #97A7B3",
"$.	c #EDECEA",
"%.	c #706F6F",
"&.	c #6D6D6B",
"*.	c #6B6B69",
"=.	c #6B6A68",
"-.	c #6A6A68",
";.	c #E0DFDB",
">.	c #A19C90",
",.	c #EDEDEB",
"'.	c #BEBCC7",
").	c #EDDFB1",
"!.	c #97A6B3",
"~.	c #96A6B2",
"{.	c #E9E8E5",
"].	c #E6E4E1",
"^.	c #E3E2DE",
"/.	c #DFDEDA",
"(.	c #DEDDD9",
"_.	c #DDDCD8",
":.	c #A19B90",
"<.	c #95A5B1",
"[.	c #BCBAC5",
"}.	c #EBDEAF",
"|.	c #95A4B1",
"1.	c #6D6C6B",
"2.	c #6C6B69",
"3.	c #6A6967",
"4.	c #696866",
"5.	c #686866",
"6.	c #DCDBD7",
"7.	c #94A3B0",
"8.	c #BBB9C4",
"9.	c #EADDAE",
"0.	c #E9DCAE",
"a.	c #94A3AF",
"b.	c #E5E4E0",
"c.	c #E4E3DF",
"d.	c #E2E1DD",
"e.	c #DBD9D5",
"f.	c #D9D7D3",
"g.	c #9F998D",
"h.	c #93A2AE",
"i.	c #BAB7C2",
"j.	c #BAB8C2",
"k.	c #92A1AD",
"l.	c #686766",
"m.	c #686765",
"n.	c #666664",
"o.	c #D8D6D2",
"p.	c #9E988D",
"q.	c #EDEDED",
"r.	c #E1E0DD",
"s.	c #E1E0DC",
"t.	c #E0DEDA",
"u.	c #DDDBD7",
"v.	c #DCDAD6",
"w.	c #D8D6D1",
"x.	c #D7D5D1",
"y.	c #9D978B",
"z.	c #D1D1D1",
"A.	c #E1DFDB",
"B.	c #DEDCD8",
"C.	c #D7D6D1",
"D.	c #D5D3CE",
"E.	c #9B958A",
"F.	c #6E6E6D",
"G.	c #999891",
"H.	c #A39E92",
"I.	c #A39D92",
"J.	c #A39D91",
"K.	c #A29C90",
"L.	c #A19B8F",
"M.	c #9D978C",
"N.	c #9B968A",
"O.	c #676359",
"P.	c #E6E5E2",
"Q.	c #91908E",
"R.	c #62615F",
"S.	c #595856",
"T.	c #575755",
"U.	c #575654",
"V.	c #565654",
"W.	c #B5B4B1",
"X.	c #B4B2AF",
"Y.	c #B3B1AE",
"Z.	c #B1AFAC",
"`.	c #B0AEAB",
" +	c #807B73",
".+	c #D8D8D4",
"++	c #D9D8D4",
"@+	c #D8D7D3",
"#+	c #D7D6D2",
"$+	c #D5D3CF",
"%+	c #D4D2CE",
"&+	c #D1CFCB",
"*+	c #D0CEC9",
"=+	c #CFCDC9",
"-+	c #979186",
";+	c #E9E8E6",
">+	c #E3E2DF",
",+	c #8F8E8C",
"'+	c #CDCCC9",
")+	c #BBBAB6",
"!+	c #B7B6B3",
"~+	c #B6B5B2",
"{+	c #B5B5B1",
"]+	c #B6B5B1",
"^+	c #B0AEAA",
"/+	c #AFADAA",
"(+	c #807B71",
"_+	c #D6D5D1",
":+	c #D6D4D0",
"<+	c #D0CECA",
"[+	c #CFCEC9",
"}+	c #CDCBC6",
"|+	c #958F85",
"                                                                                                ",
"                                                                                                ",
"                                                                                                ",
"                                                                                                ",
"                                                                                                ",
"                                                                                                ",
"                                                                                                ",
"                                                                                                ",
"                                      . . . . . . . . . . . . . . . .                           ",
"                                    . + @ # # # # # # # # $ % % & * = .                         ",
"                                    . @ # # # # # # # # # # % - ; > # , .                       ",
"                                    . # # ' ' ' ' ' ) ' ) ) % ! ~ { # ] ^ .                     ",
"                                    . # # # # # / ( _ ( _ : % < [ . . . . .                     ",
"                            . . . . . # # ' ) ) } } | } | 1 % ] 2 3 4 5 6 .                     ",
"                          . + @ # # . # / ( _ : 7 % 8 % 8 9 % 0 a b c d e .                     ",
"                          . @ # # # . # _ } | 1 f g h g h i j k l l m n o .                     ",
"                          . # # ' ' . # 7 % 8 9 p q r q r j j j j s t t u .                     ",
"                  . . . . . # # # # . # 8 g h i v w w w w x y z z A B t C .                     ",
"                . + @ # # . # # ' ) . # D E j j F G H G H I J K L M N t C .                     ",
"                . @ # # # . # / ( _ . # j O x y y P y P y Q R S T U V t C .                     ",
"                . # # ' ' . # _ } | . # j 0 G W X I l I l Y K Z ` N  .t C .                     ",
"                . # # # # . # 7 % 8 . # W ....+.@.#.$.%.%.R S &.*.=.-.;.>..                     ",
"                . # # ' ) . # 8 g h . # ,.#.'.).!.~.{.s {.{.` ].^./.(._.:..                     ",
"                . # / ( _ . # D E j . # s <.[.}.}.|.L S &.1.2.*.3.4.5.6.:..                     ",
"                . # _ } | . # j O x . # L 7.8.9.0.a.b.].b.c.d.;.(.6.e.f.g..                     ",
"                . # 7 % 8 . # j 0 G . $ N h.i.i.j.k.^.=.*.-.-.4.l.m.n.o.p..                     ",
"                . # 8 g h . # W y y . q.r.t s.;.d.;.s.;.s./.t.u.v.f.w.x.y..                     ",
"                . # D E j . # ,.$.K . z.A.t./.(.;.(.t.(.t.B.B.v.v.o.C.D.E..                     ",
"                . # j O x . # s F.F.. G.H.C I.J.I.>.K.>.K.L.:.g.g.M.M.N.O..                     ",
"                . # j 0 G . # L P.P.Q.. . . . . . . . . . . . . . . . . .                       ",
"                . # W X X . $ N *.*.3.R.S.T.U.T.V.V.W.X.Y.Z.`. +.                               ",
"                . # ,.%.R . q.r.t s.;.(..+++@+++#+o.$+%+&+*+=+-+.                               ",
"                . # s ;+;+. z.A.t./.(.;.(.t.(.t.B.B.v.v.o.C.D.E..                               ",
"                . # L 1.1.. G.H.C I.J.I.>.K.>.K.L.:.g.g.M.M.N.O..                               ",
"                . $ N >+^.,+. . . . . . . . . . . . . . . . . .                                 ",
"                . q.r.t s.6.'+)+!+~+!+{+]+X.Y.Z.^+/+(+.                                         ",
"                . z.A.t./.(.6.#+o._+o.:+:+%+%+<+[+}+|+.                                         ",
"                . G.H.C I.J.I.>.K.>.K.L.:.g.g.M.M.N.O..                                         ",
"                  . . . . . . . . . . . . . . . . . .                                           ",
"                                                                                                ",
"                                                                                                ",
"                                                                                                ",
"                                                                                                ",
"                                                                                                ",
"                                                                                                ",
"                                                                                                ",
"                                                                                                ",
"                                                                                                "};



#define DIR_LIST_WIDTH   180
#define DIR_LIST_HEIGHT  180
#define FILE_LIST_WIDTH  180
#define FILE_LIST_HEIGHT 180

/* The Hurd doesn't define either PATH_MAX or MAXPATHLEN, so we put this
 * in here, since the rest of the code in the file does require some
 * fixed maximum.
 */
#ifndef MAXPATHLEN
#  ifdef PATH_MAX
#    define MAXPATHLEN PATH_MAX
#  else
#    define MAXPATHLEN 2048
#  endif
#endif

/* I've put this here so it doesn't get confused with the 
 * file completion interface */
typedef struct _HistoryCallbackArg HistoryCallbackArg;

struct _HistoryCallbackArg
{
  gchar *directory;
  GtkWidget *menu_item;
};


typedef struct _CompletionState    CompletionState;
typedef struct _CompletionDir      CompletionDir;
typedef struct _CompletionDirSent  CompletionDirSent;
typedef struct _CompletionDirEntry CompletionDirEntry;
typedef struct _CompletionUserDir  CompletionUserDir;
typedef struct _PossibleCompletion PossibleCompletion;

/* Non-external file completion decls and structures */

/* A contant telling PRCS how many directories to cache.  Its actually
 * kept in a list, so the geometry isn't important. */
#define CMPL_DIRECTORY_CACHE_SIZE 10

/* A constant used to determine whether a substring was an exact
 * match by first_diff_index()
 */
#define PATTERN_MATCH -1
/* The arguments used by all fnmatch() calls below
 */
#define FNMATCH_FLAGS (FNM_PATHNAME | FNM_PERIOD)

#define CMPL_ERRNO_TOO_LONG ((1<<16)-1)

/* This structure contains all the useful information about a directory
 * for the purposes of filename completion.  These structures are cached
 * in the CompletionState struct.  CompletionDir's are reference counted.
 */
struct _CompletionDirSent
{
  ino_t inode;
  time_t mtime;
  dev_t device;

  gint entry_count;
  gchar *name_buffer; /* memory segment containing names of all entries */

  struct _CompletionDirEntry *entries;
};

struct _CompletionDir
{
  CompletionDirSent *sent;

  gchar *fullname;
  gint fullname_len;

  struct _CompletionDir *cmpl_parent;
  gint cmpl_index;
  gchar *cmpl_text;
};

/* This structure contains pairs of directory entry names with a flag saying
 * whether or not they are a valid directory.  NOTE: This information is used
 * to provide the caller with information about whether to update its completions
 * or try to open a file.  Since directories are cached by the directory mtime,
 * a symlink which points to an invalid file (which will not be a directory),
 * will not be reevaluated if that file is created, unless the containing
 * directory is touched.  I consider this case to be worth ignoring (josh).
 */
struct _CompletionDirEntry
{
  gint is_dir;
  gchar *entry_name;
};

struct _CompletionUserDir
{
  gchar *login;
  gchar *homedir;
};

struct _PossibleCompletion
{
  /* accessible fields, all are accessed externally by functions
   * declared above
   */
  gchar *text;
  gint is_a_completion;
  gint is_directory;

  /* Private fields
   */
  gint text_alloc;
};

struct _CompletionState
{
  gint last_valid_char;
  gchar *updated_text;
  gint updated_text_len;
  gint updated_text_alloc;
  gint re_complete;

  gchar *user_dir_name_buffer;
  gint user_directories_len;

  gchar *last_completion_text;

  gint user_completion_index; /* if >= 0, currently completing ~user */

  struct _CompletionDir *completion_dir; /* directory completing from */
  struct _CompletionDir *active_completion_dir;

  struct _PossibleCompletion the_completion;

  struct _CompletionDir *reference_dir; /* initial directory */

  GList* directory_storage;
  GList* directory_sent_storage;

  struct _CompletionUserDir *user_directories;
};


/* File completion functions which would be external, were they used
 * outside of this file.
 */

static CompletionState*    cmpl_init_state        (void);
static void                cmpl_free_state        (CompletionState *cmpl_state);
static gint                cmpl_state_okay        (CompletionState* cmpl_state);
static gchar*              cmpl_strerror          (gint);

static PossibleCompletion* cmpl_completion_matches(gchar           *text_to_complete,
						   gchar          **remaining_text,
						   CompletionState *cmpl_state);

/* Returns a name for consideration, possibly a completion, this name
 * will be invalid after the next call to cmpl_next_completion.
 */
static char*               cmpl_this_completion   (PossibleCompletion*);

/* True if this completion matches the given text.  Otherwise, this
 * output can be used to have a list of non-completions.
 */
static gint                cmpl_is_a_completion   (PossibleCompletion*);

/* True if the completion is a directory
 */
static gint                cmpl_is_directory      (PossibleCompletion*);

/* Obtains the next completion, or NULL
 */
static PossibleCompletion* cmpl_next_completion   (CompletionState*);

/* Updating completions: the return value of cmpl_updated_text() will
 * be text_to_complete completed as much as possible after the most
 * recent call to cmpl_completion_matches.  For the present
 * application, this is the suggested replacement for the user's input
 * string.  You must CALL THIS AFTER ALL cmpl_text_completions have
 * been received.
 */
static gchar*              cmpl_updated_text       (CompletionState* cmpl_state);

/* After updating, to see if the completion was a directory, call
 * this.  If it was, you should consider re-calling completion_matches.
 */
static gint                cmpl_updated_dir        (CompletionState* cmpl_state);

/* Current location: if using file completion, return the current
 * directory, from which file completion begins.  More specifically,
 * the cwd concatenated with all exact completions up to the last
 * directory delimiter('/').
 */
static gchar*              cmpl_reference_position (CompletionState* cmpl_state);

/* backing up: if cmpl_completion_matches returns NULL, you may query
 * the index of the last completable character into cmpl_updated_text.
 */
static gint                cmpl_last_valid_char    (CompletionState* cmpl_state);

/* When the user selects a non-directory, call cmpl_completion_fullname
 * to get the full name of the selected file.
 */
static gchar*              cmpl_completion_fullname (gchar*, CompletionState* cmpl_state);


/* Directory operations. */
static CompletionDir* open_ref_dir         (gchar* text_to_complete,
					    gchar** remaining_text,
					    CompletionState* cmpl_state);
static gboolean       check_dir            (gchar *dir_name, 
					    struct stat *result, 
					    gboolean *stat_subdirs);
static CompletionDir* open_dir             (gchar* dir_name,
					    CompletionState* cmpl_state);
static CompletionDir* open_user_dir        (gchar* text_to_complete,
					    CompletionState *cmpl_state);
static CompletionDir* open_relative_dir    (gchar* dir_name, CompletionDir* dir,
					    CompletionState *cmpl_state);
static CompletionDirSent* open_new_dir     (gchar* dir_name, 
					    struct stat* sbuf,
					    gboolean stat_subdirs);
static gint           correct_dir_fullname (CompletionDir* cmpl_dir);
static gint           correct_parent       (CompletionDir* cmpl_dir,
					    struct stat *sbuf);
static gchar*         find_parent_dir_fullname    (gchar* dirname);
static CompletionDir* attach_dir           (CompletionDirSent* sent,
					    gchar* dir_name,
					    CompletionState *cmpl_state);
static void           free_dir_sent (CompletionDirSent* sent);
static void           free_dir      (CompletionDir  *dir);
static void           prune_memory_usage(CompletionState *cmpl_state);

/* Completion operations */
static PossibleCompletion* attempt_homedir_completion(gchar* text_to_complete,
						      CompletionState *cmpl_state);
static PossibleCompletion* attempt_file_completion(CompletionState *cmpl_state);
static CompletionDir* find_completion_dir(gchar* text_to_complete,
					  gchar** remaining_text,
					  CompletionState* cmpl_state);
static PossibleCompletion* append_completion_text(gchar* text,
						  CompletionState* cmpl_state);
static gint get_pwdb(CompletionState* cmpl_state);
static gint first_diff_index(gchar* pat, gchar* text);
static gint compare_user_dir(const void* a, const void* b);
static gint compare_cmpl_dir(const void* a, const void* b);
static void update_cmpl(PossibleCompletion* poss,
			CompletionState* cmpl_state);

static void gtk_file_selection_class_init    (GtkFileSelectionClass *klass);
static void gtk_file_selection_init          (GtkFileSelection      *filesel);
static void gtk_file_selection_destroy       (GtkObject             *object);
static gint gtk_file_selection_key_press     (GtkWidget             *widget,
					      GdkEventKey           *event,
					      gpointer               user_data);

static void gtk_file_selection_file_button (GtkWidget *widget,
					    gint row, 
					    gint column, 
					    GdkEventButton *bevent,
					    gpointer user_data);

static void gtk_file_selection_dir_button (GtkWidget *widget,
					   gint row, 
					   gint column, 
					   GdkEventButton *bevent,
					   gpointer data);

static void gtk_file_selection_populate      (GtkFileSelection      *fs,
					      gchar                 *rel_path,
					      gint                   try_complete);
static void gtk_file_selection_abort         (GtkFileSelection      *fs);

static void gtk_file_selection_update_history_menu (GtkFileSelection       *fs,
						    gchar                  *current_dir);

static void gtk_file_selection_create_dir (GtkWidget *widget, gpointer data);
static void gtk_file_selection_delete_file (GtkWidget *widget, gpointer data);
static void gtk_file_selection_rename_file (GtkWidget *widget, gpointer data);

static void home_clicked (GtkWidget *widget, gpointer data);
static void desktop_clicked (GtkWidget *widget, gpointer data);
static void documents_clicked (GtkWidget *widget, gpointer data);


static GtkWindowClass *parent_class = NULL;

/* Saves errno when something cmpl does fails. */
static gint cmpl_errno;

GtkType
gtk_file_selection_get_type (void)
{
  static GtkType file_selection_type = 0;

  if (!file_selection_type)
    {
      static const GtkTypeInfo filesel_info =
      {
	"GtkFileSelection",
	sizeof (GtkFileSelection),
	sizeof (GtkFileSelectionClass),
	(GtkClassInitFunc) gtk_file_selection_class_init,
	(GtkObjectInitFunc) gtk_file_selection_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL,
      };

      file_selection_type = gtk_type_unique (GTK_TYPE_WINDOW, &filesel_info);
    }

  return file_selection_type;
}

static void
gtk_file_selection_class_init (GtkFileSelectionClass *class)
{
  GtkObjectClass *object_class;

  object_class = (GtkObjectClass*) class;

  parent_class = gtk_type_class (GTK_TYPE_WINDOW);

  object_class->destroy = gtk_file_selection_destroy;
}

static GtkWidget *
create_pixmap_button (GtkFileSelection *filesel, 
		      GtkWidget *parent_hbox,
		      gchar **xpm_data,
		      gchar *title)
{
  GtkWidget *btn;
  GtkWidget *hbox;
  GtkWidget *wpixmap;
  GtkWidget *label;
  GdkPixmap *pixmap;
  GdkPixmap *mask;
  
  btn = gtk_button_new ();
  gtk_box_pack_start (GTK_BOX (parent_hbox), btn, TRUE, FALSE, 5);
  gtk_widget_realize (btn);

  pixmap = gdk_pixmap_create_from_xpm_d (btn->window, &mask, 
					 &btn->style->bg[GTK_STATE_NORMAL], 
					 (gchar**)xpm_data);
  
  label = gtk_label_new (title);
  
  wpixmap = gtk_pixmap_new (pixmap, mask);
  
  hbox = gtk_vbox_new (FALSE, 5);

  gtk_box_pack_start (GTK_BOX (hbox), wpixmap, FALSE, FALSE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);

  gtk_widget_show_all (hbox);

  gtk_container_add (GTK_CONTAINER (btn), hbox);

  return btn;
}

static void
gtk_file_selection_init (GtkFileSelection *filesel)
{
  GtkWidget *entry_vbox;
  GtkWidget *label;
  GtkWidget *list_hbox;
  GtkWidget *confirm_area;
  GtkWidget *pulldown_hbox;
  GtkWidget *scrolled_win;
  GtkWidget *navigation_area;
  GtkWidget *btn;

  char *dirname;
  char *dir_title [2];
  char *file_title [2];
  
  filesel->cmpl_state = cmpl_init_state ();

  /* The dialog-sized vertical box  */
  filesel->main_vbox = gtk_vbox_new (FALSE, 10);
  gtk_container_set_border_width (GTK_CONTAINER (filesel), 10);
  gtk_container_add (GTK_CONTAINER (filesel), filesel->main_vbox);
  gtk_widget_show (filesel->main_vbox);

  /* The horizontal box containing create, rename etc. buttons */
  filesel->button_area = gtk_hbutton_box_new ();
  gtk_button_box_set_layout(GTK_BUTTON_BOX(filesel->button_area), GTK_BUTTONBOX_START);
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(filesel->button_area), 0);
  gtk_box_pack_start (GTK_BOX (filesel->main_vbox), filesel->button_area, 
		      FALSE, FALSE, 0);
  gtk_widget_show (filesel->button_area);
  
  gtk_file_selection_show_fileop_buttons(filesel);

  /* hbox for pulldown menu */
  pulldown_hbox = gtk_hbox_new (TRUE, 5);
  gtk_box_pack_start (GTK_BOX (filesel->main_vbox), pulldown_hbox, FALSE, FALSE, 0);
  gtk_widget_show (pulldown_hbox);
  
  /* Pulldown menu */
  filesel->history_pulldown = gtk_option_menu_new ();
  gtk_widget_show (filesel->history_pulldown);
  gtk_box_pack_start (GTK_BOX (pulldown_hbox), filesel->history_pulldown, 
		      FALSE, FALSE, 0);
    
  /*  The horizontal box containing the directory and file listboxes  */
  list_hbox = gtk_hbox_new (FALSE, 5);
  gtk_box_pack_start (GTK_BOX (filesel->main_vbox), list_hbox, TRUE, TRUE, 0);
  gtk_widget_show (list_hbox);

  /* The navigation buttons */
  navigation_area = gtk_vbox_new (TRUE, 0);
  gtk_box_pack_start (GTK_BOX (list_hbox), navigation_area, FALSE, FALSE, 5);
  gtk_widget_show (navigation_area);

  if (g_get_home_dir () != NULL) 
    {
      struct stat sb;
      btn = create_pixmap_button (filesel, navigation_area, 
				  (gchar **)home_xpm, 
				  _("Home"));
      gtk_widget_show (btn);
      gtk_signal_connect (GTK_OBJECT (btn), "clicked",
			  home_clicked, filesel);

      dirname = g_strdup_printf ("%s/Desktop/", g_get_home_dir ());
      if (stat (dirname, &sb) == 0 && S_ISDIR(sb.st_mode))
	{
	  btn = create_pixmap_button (filesel, navigation_area, 
				      (gchar **)desktop_xpm, 
				      _("Desktop"));
	  gtk_widget_show (btn);
	  gtk_signal_connect (GTK_OBJECT (btn), "clicked",
			      desktop_clicked, filesel);

	  
	}
      g_free (dirname);
      
      dirname = g_strdup_printf ("%s/Documents/", g_get_home_dir ());
      if (stat (dirname, &sb) == 0 && S_ISDIR(sb.st_mode))
	{
	  btn = create_pixmap_button (filesel, navigation_area, 
				      (gchar **)documents_xpm, 
				      _("Documents"));
	  gtk_widget_show (btn);
	  gtk_signal_connect (GTK_OBJECT (btn), "clicked",
			      documents_clicked, filesel);
	}
      
      g_free (dirname);
    }

  /* The directories clist */
  dir_title[0] = _("Directories");
  dir_title[1] = NULL;
  filesel->dir_list = gtk_clist_new_with_titles (1, (gchar**) dir_title);
  gtk_widget_set_usize (filesel->dir_list, DIR_LIST_WIDTH, DIR_LIST_HEIGHT);
  gtk_signal_connect (GTK_OBJECT (filesel->dir_list), "select_row",
		      (GtkSignalFunc) gtk_file_selection_dir_button, 
		      (gpointer) filesel);
  gtk_clist_column_titles_passive (GTK_CLIST (filesel->dir_list));

  scrolled_win = gtk_scrolled_window_new (NULL, NULL);
  gtk_container_add (GTK_CONTAINER (scrolled_win), filesel->dir_list);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_win),
				  GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_container_set_border_width (GTK_CONTAINER (scrolled_win), 5);
  gtk_box_pack_start (GTK_BOX (list_hbox), scrolled_win, TRUE, TRUE, 0);
  gtk_widget_show (filesel->dir_list);
  gtk_widget_show (scrolled_win);

  /* The files clist */
  file_title[0] = _("Files");
  file_title[1] = NULL;
  filesel->file_list = gtk_clist_new_with_titles (1, (gchar**) file_title);
  gtk_widget_set_usize (filesel->file_list, FILE_LIST_WIDTH, FILE_LIST_HEIGHT);
  gtk_signal_connect (GTK_OBJECT (filesel->file_list), "select_row",
		      (GtkSignalFunc) gtk_file_selection_file_button, 
		      (gpointer) filesel);
  gtk_clist_column_titles_passive (GTK_CLIST (filesel->file_list));

  scrolled_win = gtk_scrolled_window_new (NULL, NULL);
  gtk_container_add (GTK_CONTAINER (scrolled_win), filesel->file_list);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_win),
				  GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_container_set_border_width (GTK_CONTAINER (scrolled_win), 5);
  gtk_box_pack_start (GTK_BOX (list_hbox), scrolled_win, TRUE, TRUE, 0);
  gtk_widget_show (filesel->file_list);
  gtk_widget_show (scrolled_win);

  /* action area for packing buttons into. */
  filesel->action_area = gtk_hbox_new (TRUE, 0);
  gtk_box_pack_start (GTK_BOX (filesel->main_vbox), filesel->action_area, 
		      FALSE, FALSE, 0);
  gtk_widget_show (filesel->action_area);
  
  /*  The OK/Cancel button area */
  confirm_area = gtk_hbutton_box_new ();
  gtk_button_box_set_layout(GTK_BUTTON_BOX(confirm_area), GTK_BUTTONBOX_END);
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(confirm_area), 5);
  gtk_box_pack_end (GTK_BOX (filesel->main_vbox), confirm_area, FALSE, FALSE, 0);
  gtk_widget_show (confirm_area);

  /*  The OK button  */
  filesel->ok_button = gtk_button_new_with_label (_("OK"));
  GTK_WIDGET_SET_FLAGS (filesel->ok_button, GTK_CAN_DEFAULT);
  gtk_box_pack_start (GTK_BOX (confirm_area), filesel->ok_button, TRUE, TRUE, 0);
  gtk_widget_grab_default (filesel->ok_button);
  gtk_widget_show (filesel->ok_button);

  /*  The Cancel button  */
  filesel->cancel_button = gtk_button_new_with_label (_("Cancel"));
  GTK_WIDGET_SET_FLAGS (filesel->cancel_button, GTK_CAN_DEFAULT);
  gtk_box_pack_start (GTK_BOX (confirm_area), filesel->cancel_button, TRUE, TRUE, 0);
  gtk_widget_show (filesel->cancel_button);

  /*  The selection entry widget  */
  entry_vbox = gtk_vbox_new (FALSE, 2);
  gtk_box_pack_end (GTK_BOX (filesel->main_vbox), entry_vbox, FALSE, FALSE, 0);
  gtk_widget_show (entry_vbox);

  filesel->selection_text = label = gtk_label_new ("");
  gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
  gtk_box_pack_start (GTK_BOX (entry_vbox), label, FALSE, FALSE, 0);
  gtk_widget_show (label);

  filesel->selection_entry = gtk_entry_new ();
  gtk_signal_connect (GTK_OBJECT (filesel->selection_entry), "key_press_event",
		      (GtkSignalFunc) gtk_file_selection_key_press, filesel);
  gtk_signal_connect_object (GTK_OBJECT (filesel->selection_entry), "focus_in_event",
			     (GtkSignalFunc) gtk_widget_grab_default,
			     GTK_OBJECT (filesel->ok_button));
  gtk_signal_connect_object (GTK_OBJECT (filesel->selection_entry), "activate",
                             (GtkSignalFunc) gtk_button_clicked,
                             GTK_OBJECT (filesel->ok_button));
  gtk_box_pack_start (GTK_BOX (entry_vbox), filesel->selection_entry, TRUE, TRUE, 0);
  gtk_object_set_data (GTK_OBJECT (filesel->selection_entry), "last_entry_was_dir_selection", (gpointer) FALSE);
  gtk_widget_show (filesel->selection_entry);

  if (!cmpl_state_okay (filesel->cmpl_state))
    {
      gchar err_buf[256];

      sprintf (err_buf, _("Directory unreadable: %s"), cmpl_strerror (cmpl_errno));

      gtk_label_set_text (GTK_LABEL (filesel->selection_text), err_buf);
    }
  else
    {
      gtk_file_selection_populate (filesel, "", FALSE);
    }

  gtk_widget_grab_focus (filesel->selection_entry);
}

GtkWidget*
gtk_file_selection_new (const gchar *title)
{
  GtkFileSelection *filesel;

  filesel = gtk_type_new (GTK_TYPE_FILE_SELECTION);
  gtk_window_set_title (GTK_WINDOW (filesel), title);

  return GTK_WIDGET (filesel);
}

void
gtk_file_selection_show_fileop_buttons (GtkFileSelection *filesel)
{
  g_return_if_fail (filesel != NULL);
  g_return_if_fail (GTK_IS_FILE_SELECTION (filesel));
    
  /* delete, create directory, and rename */
  if (!filesel->fileop_c_dir) 
    {
      filesel->fileop_c_dir = gtk_button_new_with_label (_("Create Dir"));
      gtk_signal_connect (GTK_OBJECT (filesel->fileop_c_dir), "clicked",
			  (GtkSignalFunc) gtk_file_selection_create_dir, 
			  (gpointer) filesel);
      gtk_box_pack_start (GTK_BOX (filesel->button_area), 
			  filesel->fileop_c_dir, TRUE, TRUE, 0);
      gtk_widget_show (filesel->fileop_c_dir);
    }
	
  if (!filesel->fileop_del_file) 
    {
      filesel->fileop_del_file = gtk_button_new_with_label (_("Delete File"));
      gtk_signal_connect (GTK_OBJECT (filesel->fileop_del_file), "clicked",
			  (GtkSignalFunc) gtk_file_selection_delete_file, 
			  (gpointer) filesel);
      gtk_box_pack_start (GTK_BOX (filesel->button_area), 
			  filesel->fileop_del_file, TRUE, TRUE, 0);
      gtk_widget_show (filesel->fileop_del_file);
    }

  if (!filesel->fileop_ren_file)
    {
      filesel->fileop_ren_file = gtk_button_new_with_label (_("Rename File"));
      gtk_signal_connect (GTK_OBJECT (filesel->fileop_ren_file), "clicked",
			  (GtkSignalFunc) gtk_file_selection_rename_file, 
			  (gpointer) filesel);
      gtk_box_pack_start (GTK_BOX (filesel->button_area), 
			  filesel->fileop_ren_file, TRUE, TRUE, 0);
      gtk_widget_show (filesel->fileop_ren_file);
    }

  gtk_widget_queue_resize(GTK_WIDGET(filesel));
}

void       
gtk_file_selection_hide_fileop_buttons (GtkFileSelection *filesel)
{
  g_return_if_fail (filesel != NULL);
  g_return_if_fail (GTK_IS_FILE_SELECTION (filesel));
    
  if (filesel->fileop_ren_file) 
    {
      gtk_widget_destroy (filesel->fileop_ren_file);
      filesel->fileop_ren_file = NULL;
    }

  if (filesel->fileop_del_file)
    {
      gtk_widget_destroy (filesel->fileop_del_file);
      filesel->fileop_del_file = NULL;
    }

  if (filesel->fileop_c_dir)
    {
      gtk_widget_destroy (filesel->fileop_c_dir);
      filesel->fileop_c_dir = NULL;
    }
}



void
gtk_file_selection_set_filename (GtkFileSelection *filesel,
				 const gchar      *filename)
{
  char  buf[MAXPATHLEN];
  const char *name, *last_slash;

  g_return_if_fail (filesel != NULL);
  g_return_if_fail (GTK_IS_FILE_SELECTION (filesel));
  g_return_if_fail (filename != NULL);

  last_slash = strrchr (filename, '/');

  if (!last_slash)
    {
      buf[0] = 0;
      name = filename;
    }
  else
    {
      gint len = MIN (MAXPATHLEN - 1, last_slash - filename + 1);

      strncpy (buf, filename, len);
      buf[len] = 0;

      name = last_slash + 1;
    }

  gtk_file_selection_populate (filesel, buf, FALSE);

  if (filesel->selection_entry)
    gtk_entry_set_text (GTK_ENTRY (filesel->selection_entry), name);
}

gchar*
gtk_file_selection_get_filename (GtkFileSelection *filesel)
{
  static char nothing[2] = "";
  char *text;
  char *filename;

  g_return_val_if_fail (filesel != NULL, nothing);
  g_return_val_if_fail (GTK_IS_FILE_SELECTION (filesel), nothing);

  text = gtk_entry_get_text (GTK_ENTRY (filesel->selection_entry));
  if (text)
    {
      filename = cmpl_completion_fullname (text, filesel->cmpl_state);
      return filename;
    }

  return nothing;
}

void
gtk_file_selection_complete (GtkFileSelection *filesel,
			     const gchar      *pattern)
{
  g_return_if_fail (filesel != NULL);
  g_return_if_fail (GTK_IS_FILE_SELECTION (filesel));
  g_return_if_fail (pattern != NULL);

  if (filesel->selection_entry)
    gtk_entry_set_text (GTK_ENTRY (filesel->selection_entry), pattern);
  gtk_file_selection_populate (filesel, (gchar*) pattern, TRUE);
}

static void
gtk_file_selection_destroy (GtkObject *object)
{
  GtkFileSelection *filesel;
  GList *list;
  HistoryCallbackArg *callback_arg;

  g_return_if_fail (object != NULL);
  g_return_if_fail (GTK_IS_FILE_SELECTION (object));

  filesel = GTK_FILE_SELECTION (object);
  
  if (filesel->fileop_dialog)
	  gtk_widget_destroy (filesel->fileop_dialog);
  
  if (filesel->history_list)
    {
      list = filesel->history_list;
      while (list)
	{
	  callback_arg = list->data;
	  g_free (callback_arg->directory);
	  g_free (callback_arg);
	  list = list->next;
	}
      g_list_free (filesel->history_list);
      filesel->history_list = NULL;
    }
  
  cmpl_free_state (filesel->cmpl_state);
  filesel->cmpl_state = NULL;

  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

/* Begin file operations callbacks */

static void
gtk_file_selection_fileop_error (GtkFileSelection *fs, gchar *error_message)
{
  GtkWidget *label;
  GtkWidget *vbox;
  GtkWidget *button;
  GtkWidget *dialog;
  
  g_return_if_fail (error_message != NULL);
  
  /* main dialog */
  dialog = gtk_dialog_new ();
  /*
  gtk_signal_connect (GTK_OBJECT (dialog), "destroy",
		      (GtkSignalFunc) gtk_file_selection_fileop_destroy, 
		      (gpointer) fs);
  */
  gtk_window_set_title (GTK_WINDOW (dialog), _("Error"));
  gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
  
  /* If file dialog is grabbed, make this dialog modal too */
  /* When error dialog is closed, file dialog will be grabbed again */
  if (GTK_WINDOW(fs)->modal)
      gtk_window_set_modal (GTK_WINDOW(dialog), TRUE);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 8);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), vbox,
		     FALSE, FALSE, 0);
  gtk_widget_show(vbox);

  label = gtk_label_new(error_message);
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 5);
  gtk_widget_show(label);

  /* yes, we free it */
  g_free (error_message);
  
  /* close button */
  button = gtk_button_new_with_label (_("Close"));
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     (GtkSignalFunc) gtk_widget_destroy, 
			     (gpointer) dialog);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),
		     button, TRUE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_widget_grab_default(button);
  gtk_widget_show (button);

  gtk_widget_show (dialog);
}

static void
gtk_file_selection_fileop_destroy (GtkWidget *widget, gpointer data)
{
  GtkFileSelection *fs = data;

  g_return_if_fail (fs != NULL);
  g_return_if_fail (GTK_IS_FILE_SELECTION (fs));
  
  fs->fileop_dialog = NULL;
}


static void
gtk_file_selection_create_dir_confirmed (GtkWidget *widget, gpointer data)
{
  GtkFileSelection *fs = data;
  gchar *dirname;
  gchar *path;
  gchar *full_path;
  gchar *buf;
  CompletionState *cmpl_state;
  
  g_return_if_fail (fs != NULL);
  g_return_if_fail (GTK_IS_FILE_SELECTION (fs));

  dirname = gtk_entry_get_text (GTK_ENTRY (fs->fileop_entry));
  cmpl_state = (CompletionState*) fs->cmpl_state;
  path = cmpl_reference_position (cmpl_state);
  
  full_path = g_strconcat (path, "/", dirname, NULL);
  if ( (mkdir (full_path, 0755) < 0) ) 
    {
      buf = g_strconcat ("Error creating directory \"", dirname, "\":  ", 
			 g_strerror(errno), NULL);
      gtk_file_selection_fileop_error (fs, buf);
    }
  g_free (full_path);
  
  gtk_widget_destroy (fs->fileop_dialog);
  gtk_file_selection_populate (fs, "", FALSE);
}
  
static void
gtk_file_selection_create_dir (GtkWidget *widget, gpointer data)
{
  GtkFileSelection *fs = data;
  GtkWidget *label;
  GtkWidget *dialog;
  GtkWidget *vbox;
  GtkWidget *button;

  g_return_if_fail (fs != NULL);
  g_return_if_fail (GTK_IS_FILE_SELECTION (fs));

  if (fs->fileop_dialog)
	  return;
  
  /* main dialog */
  fs->fileop_dialog = dialog = gtk_dialog_new ();
  gtk_signal_connect (GTK_OBJECT (dialog), "destroy",
		      (GtkSignalFunc) gtk_file_selection_fileop_destroy, 
		      (gpointer) fs);
  gtk_window_set_title (GTK_WINDOW (dialog), _("Create Directory"));
  gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);

  /* If file dialog is grabbed, grab option dialog */
  /* When option dialog is closed, file dialog will be grabbed again */
  if (GTK_WINDOW(fs)->modal)
      gtk_window_set_modal (GTK_WINDOW(dialog), TRUE);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 8);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), vbox,
		     FALSE, FALSE, 0);
  gtk_widget_show(vbox);
  
  label = gtk_label_new(_("Directory name:"));
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 5);
  gtk_widget_show(label);

  /*  The directory entry widget  */
  fs->fileop_entry = gtk_entry_new ();
  gtk_box_pack_start (GTK_BOX (vbox), fs->fileop_entry, 
		      TRUE, TRUE, 5);
  GTK_WIDGET_SET_FLAGS(fs->fileop_entry, GTK_CAN_DEFAULT);
  gtk_widget_show (fs->fileop_entry);
  
  /* buttons */
  button = gtk_button_new_with_label (_("Create"));
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      (GtkSignalFunc) gtk_file_selection_create_dir_confirmed, 
		      (gpointer) fs);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),
		     button, TRUE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_widget_show(button);
  
  button = gtk_button_new_with_label (_("Cancel"));
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     (GtkSignalFunc) gtk_widget_destroy, 
			     (gpointer) dialog);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),
		     button, TRUE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_widget_grab_default(button);
  gtk_widget_show (button);

  gtk_widget_show (dialog);
}

static void
gtk_file_selection_delete_file_confirmed (GtkWidget *widget, gpointer data)
{
  GtkFileSelection *fs = data;
  CompletionState *cmpl_state;
  gchar *path;
  gchar *full_path;
  gchar *buf;
  
  g_return_if_fail (fs != NULL);
  g_return_if_fail (GTK_IS_FILE_SELECTION (fs));

  cmpl_state = (CompletionState*) fs->cmpl_state;
  path = cmpl_reference_position (cmpl_state);
  
  full_path = g_strconcat (path, "/", fs->fileop_file, NULL);
  if ( (unlink (full_path) < 0) ) 
    {
      buf = g_strconcat ("Error deleting file \"", fs->fileop_file, "\":  ", 
			 g_strerror(errno), NULL);
      gtk_file_selection_fileop_error (fs, buf);
    }
  g_free (full_path);
  
  gtk_widget_destroy (fs->fileop_dialog);
  gtk_file_selection_populate (fs, "", FALSE);
}

static void
gtk_file_selection_delete_file (GtkWidget *widget, gpointer data)
{
  GtkFileSelection *fs = data;
  GtkWidget *label;
  GtkWidget *vbox;
  GtkWidget *button;
  GtkWidget *dialog;
  gchar *filename;
  gchar *buf;
  
  g_return_if_fail (fs != NULL);
  g_return_if_fail (GTK_IS_FILE_SELECTION (fs));

  if (fs->fileop_dialog)
	  return;

  filename = gtk_entry_get_text (GTK_ENTRY (fs->selection_entry));
  if (strlen(filename) < 1)
	  return;

  fs->fileop_file = filename;
  
  /* main dialog */
  fs->fileop_dialog = dialog = gtk_dialog_new ();
  gtk_signal_connect (GTK_OBJECT (dialog), "destroy",
		      (GtkSignalFunc) gtk_file_selection_fileop_destroy, 
		      (gpointer) fs);
  gtk_window_set_title (GTK_WINDOW (dialog), _("Delete File"));
  gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);

  /* If file dialog is grabbed, grab option dialog */
  /* When option dialog is closed, file dialog will be grabbed again */
  if (GTK_WINDOW(fs)->modal)
      gtk_window_set_modal (GTK_WINDOW(dialog), TRUE);
  
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 8);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), vbox,
		     FALSE, FALSE, 0);
  gtk_widget_show(vbox);

  buf = g_strconcat ("Really delete file \"", filename, "\" ?", NULL);
  label = gtk_label_new(buf);
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 5);
  gtk_widget_show(label);
  g_free(buf);
  
  /* buttons */
  button = gtk_button_new_with_label (_("Delete"));
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      (GtkSignalFunc) gtk_file_selection_delete_file_confirmed, 
		      (gpointer) fs);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),
		     button, TRUE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_widget_show(button);
  
  button = gtk_button_new_with_label (_("Cancel"));
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     (GtkSignalFunc) gtk_widget_destroy, 
			     (gpointer) dialog);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),
		     button, TRUE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_widget_grab_default(button);
  gtk_widget_show (button);

  gtk_widget_show (dialog);

}

static void
gtk_file_selection_rename_file_confirmed (GtkWidget *widget, gpointer data)
{
  GtkFileSelection *fs = data;
  gchar *buf;
  gchar *file;
  gchar *path;
  gchar *new_filename;
  gchar *old_filename;
  CompletionState *cmpl_state;
  
  g_return_if_fail (fs != NULL);
  g_return_if_fail (GTK_IS_FILE_SELECTION (fs));

  file = gtk_entry_get_text (GTK_ENTRY (fs->fileop_entry));
  cmpl_state = (CompletionState*) fs->cmpl_state;
  path = cmpl_reference_position (cmpl_state);
  
  new_filename = g_strconcat (path, "/", file, NULL);
  old_filename = g_strconcat (path, "/", fs->fileop_file, NULL);

  if ( (rename (old_filename, new_filename)) < 0) 
    {
      buf = g_strconcat ("Error renaming file \"", file, "\":  ", 
			 g_strerror(errno), NULL);
      gtk_file_selection_fileop_error (fs, buf);
    }
  g_free (new_filename);
  g_free (old_filename);
  
  gtk_widget_destroy (fs->fileop_dialog);
  gtk_file_selection_populate (fs, "", FALSE);
}
  
static void
gtk_file_selection_rename_file (GtkWidget *widget, gpointer data)
{
  GtkFileSelection *fs = data;
  GtkWidget *label;
  GtkWidget *dialog;
  GtkWidget *vbox;
  GtkWidget *button;
  gchar *buf;
  
  g_return_if_fail (fs != NULL);
  g_return_if_fail (GTK_IS_FILE_SELECTION (fs));

  if (fs->fileop_dialog)
	  return;

  fs->fileop_file = gtk_entry_get_text (GTK_ENTRY (fs->selection_entry));
  if (strlen(fs->fileop_file) < 1)
	  return;
  
  /* main dialog */
  fs->fileop_dialog = dialog = gtk_dialog_new ();
  gtk_signal_connect (GTK_OBJECT (dialog), "destroy",
		      (GtkSignalFunc) gtk_file_selection_fileop_destroy, 
		      (gpointer) fs);
  gtk_window_set_title (GTK_WINDOW (dialog), _("Rename File"));
  gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);

  /* If file dialog is grabbed, grab option dialog */
  /* When option dialog  closed, file dialog will be grabbed again */
  if (GTK_WINDOW(fs)->modal)
    gtk_window_set_modal (GTK_WINDOW(dialog), TRUE);
  
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER(vbox), 8);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), vbox,
		     FALSE, FALSE, 0);
  gtk_widget_show(vbox);
  
  buf = g_strconcat ("Rename file \"", fs->fileop_file, "\" to:", NULL);
  label = gtk_label_new(buf);
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 5);
  gtk_widget_show(label);
  g_free(buf);

  /* New filename entry */
  fs->fileop_entry = gtk_entry_new ();
  gtk_box_pack_start (GTK_BOX (vbox), fs->fileop_entry, 
		      TRUE, TRUE, 5);
  GTK_WIDGET_SET_FLAGS(fs->fileop_entry, GTK_CAN_DEFAULT);
  gtk_widget_show (fs->fileop_entry);
  
  gtk_entry_set_text (GTK_ENTRY (fs->fileop_entry), fs->fileop_file);
  gtk_editable_select_region (GTK_EDITABLE (fs->fileop_entry),
			      0, strlen (fs->fileop_file));

  /* buttons */
  button = gtk_button_new_with_label (_("Rename"));
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      (GtkSignalFunc) gtk_file_selection_rename_file_confirmed, 
		      (gpointer) fs);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),
		     button, TRUE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_widget_show(button);
  
  button = gtk_button_new_with_label (_("Cancel"));
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     (GtkSignalFunc) gtk_widget_destroy, 
			     (gpointer) dialog);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),
		     button, TRUE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_widget_grab_default(button);
  gtk_widget_show (button);

  gtk_widget_show (dialog);
}


static void
home_clicked (GtkWidget *widget, gpointer data)
{
  char *dir;
  
  dir = g_strdup_printf ("%s/", g_get_home_dir());
  
  gtk_file_selection_populate (GTK_FILE_SELECTION (data), 
			       dir, FALSE);
  g_free (dir);
}

static void
desktop_clicked (GtkWidget *widget, gpointer data)
{
  char *dir;
  
  dir = g_strdup_printf ("%s/Desktop/", g_get_home_dir ());
  
  gtk_file_selection_populate (GTK_FILE_SELECTION (data), 
			       dir, FALSE);
  g_free (dir);
}

static void 
documents_clicked (GtkWidget *widget, gpointer data)
{
  char *dir;
  dir = g_strdup_printf ("%s/Documents/", g_get_home_dir ());
  
  gtk_file_selection_populate (GTK_FILE_SELECTION (data), dir, FALSE);
  
  g_free (dir);
}

static gint
gtk_file_selection_key_press (GtkWidget   *widget,
			      GdkEventKey *event,
			      gpointer     user_data)
{
  GtkFileSelection *fs;
  char *text;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  fs = GTK_FILE_SELECTION (user_data);
  gtk_object_set_data (GTK_OBJECT (fs->selection_entry), "last_entry_was_dir_selection", (gpointer) FALSE);

  if (event->keyval == GDK_Tab)
    {
      text = gtk_entry_get_text (GTK_ENTRY (fs->selection_entry));

      text = g_strdup (text);

      gtk_file_selection_populate (fs, text, TRUE);

      g_free (text);

      gtk_signal_emit_stop_by_name (GTK_OBJECT (widget), "key_press_event");

      return TRUE;
    }

  return FALSE;
}


static void
gtk_file_selection_history_callback (GtkWidget *widget, gpointer data)
{
  GtkFileSelection *fs = data;
  HistoryCallbackArg *callback_arg;
  GList *list;

  g_return_if_fail (fs != NULL);
  g_return_if_fail (GTK_IS_FILE_SELECTION (fs));

  list = fs->history_list;
  
  while (list) {
    callback_arg = list->data;
    
    if (callback_arg->menu_item == widget)
      {
	gtk_file_selection_populate (fs, callback_arg->directory, FALSE);
	break;
      }
    
    list = list->next;
  }
}

static void 
gtk_file_selection_update_history_menu (GtkFileSelection *fs,
					gchar *current_directory)
{
  HistoryCallbackArg *callback_arg;
  GtkWidget *menu_item;
  GList *list;
  gchar *current_dir;
  gint dir_len;
  gint i;
  
  g_return_if_fail (fs != NULL);
  g_return_if_fail (GTK_IS_FILE_SELECTION (fs));
  g_return_if_fail (current_directory != NULL);
  
  list = fs->history_list;

  if (fs->history_menu) 
    {
      while (list) {
	callback_arg = list->data;
	g_free (callback_arg->directory);
	g_free (callback_arg);
	list = list->next;
      }
      g_list_free (fs->history_list);
      fs->history_list = NULL;
      
      gtk_widget_destroy (fs->history_menu);
    }
  
  fs->history_menu = gtk_menu_new();

  current_dir = g_strdup (current_directory);

  dir_len = strlen (current_dir);

  for (i = dir_len; i >= 0; i--)
    {
      /* the i == dir_len is to catch the full path for the first 
       * entry. */
      if ( (current_dir[i] == '/') || (i == dir_len))
	{
	  /* another small hack to catch the full path */
	  if (i != dir_len) 
		  current_dir[i + 1] = '\0';
	  menu_item = gtk_menu_item_new_with_label (current_dir);
	  
	  callback_arg = g_new (HistoryCallbackArg, 1);
	  callback_arg->menu_item = menu_item;
	  
	  /* since the autocompletion gets confused if you don't 
	   * supply a trailing '/' on a dir entry, set the full
	   * (current) path to "" which just refreshes the filesel */
	  if (dir_len == i) {
	    callback_arg->directory = g_strdup ("");
	  } else {
	    callback_arg->directory = g_strdup (current_dir);
	  }
	  
	  fs->history_list = g_list_append (fs->history_list, callback_arg);
	  
	  gtk_signal_connect (GTK_OBJECT (menu_item), "activate",
			      (GtkSignalFunc) gtk_file_selection_history_callback,
			      (gpointer) fs);
	  gtk_menu_append (GTK_MENU (fs->history_menu), menu_item);
	  gtk_widget_show (menu_item);
	}
    }

  gtk_option_menu_set_menu (GTK_OPTION_MENU (fs->history_pulldown), 
			    fs->history_menu);
  g_free (current_dir);
}

static void
gtk_file_selection_file_button (GtkWidget *widget,
			       gint row, 
			       gint column, 
			       GdkEventButton *bevent,
			       gpointer user_data)
{
  GtkFileSelection *fs = NULL;
  gchar *filename, *temp = NULL;
  
  g_return_if_fail (GTK_IS_CLIST (widget));

  fs = user_data;
  g_return_if_fail (fs != NULL);
  g_return_if_fail (GTK_IS_FILE_SELECTION (fs));
  
  gtk_clist_get_text (GTK_CLIST (fs->file_list), row, 0, &temp);
  filename = g_strdup (temp);

  if (filename)
    {
      if (bevent)
	switch (bevent->type)
	  {
	  case GDK_2BUTTON_PRESS:
	    gtk_button_clicked (GTK_BUTTON (fs->ok_button));
	    break;
	    
	  default:
	    gtk_entry_set_text (GTK_ENTRY (fs->selection_entry), filename);
	    break;
	  }
      else
	gtk_entry_set_text (GTK_ENTRY (fs->selection_entry), filename);

      gtk_object_set_data (GTK_OBJECT (fs->selection_entry), "last_entry_was_dir_selection", (gpointer) FALSE);
      g_free (filename);
    }
}

static void
gtk_file_selection_dir_button (GtkWidget *widget,
			       gint row, 
			       gint column, 
			       GdkEventButton *bevent,
			       gpointer user_data)
{
  GtkFileSelection *fs = NULL;
  gchar *filename, *temp = NULL;
  const gchar *entry_text;
  gboolean last_entry_was_dir_selection;

  g_return_if_fail (GTK_IS_CLIST (widget));

  fs = GTK_FILE_SELECTION (user_data);
  g_return_if_fail (fs != NULL);
  g_return_if_fail (GTK_IS_FILE_SELECTION (fs));

  gtk_clist_get_text (GTK_CLIST (fs->dir_list), row, 0, &temp);
  filename = g_strdup (temp);

  entry_text = gtk_entry_get_text (GTK_ENTRY (fs->selection_entry));
  last_entry_was_dir_selection = (gboolean) gtk_object_get_data (GTK_OBJECT (fs->selection_entry),
								 "last_entry_was_dir_selection");

  if (filename)
    {
      if (bevent)
	switch (bevent->type)
	  {
	  case GDK_2BUTTON_PRESS:
	    gtk_file_selection_populate (fs, filename, FALSE);
	    break;
	  
	  default:
	    if (!entry_text || !strlen (entry_text) || last_entry_was_dir_selection)
	      {
		gtk_entry_set_text (GTK_ENTRY (fs->selection_entry), filename);
		last_entry_was_dir_selection = TRUE;
	      }
	    break;
	  }
      else if (!entry_text || !strlen (entry_text) || last_entry_was_dir_selection)
	  {
	    gtk_entry_set_text (GTK_ENTRY (fs->selection_entry), filename);
	    last_entry_was_dir_selection = TRUE;
	  }

      g_free (filename);
    }

  gtk_object_set_data (GTK_OBJECT (fs->selection_entry), "last_entry_was_dir_selection",
		       (gpointer) last_entry_was_dir_selection);
}

static void
gtk_file_selection_populate (GtkFileSelection *fs,
			     gchar            *rel_path,
			     gint              try_complete)
{
  CompletionState *cmpl_state;
  PossibleCompletion* poss;
  gchar* filename;
  gint row;
  gchar* rem_path = rel_path;
  gchar* sel_text;
  gchar* text[2];
  gint did_recurse = FALSE;
  gint possible_count = 0;
  gint selection_index = -1;
  gint file_list_width;
  gint dir_list_width;
  
  g_return_if_fail (fs != NULL);
  g_return_if_fail (GTK_IS_FILE_SELECTION (fs));
  
  cmpl_state = (CompletionState*) fs->cmpl_state;
  poss = cmpl_completion_matches (rel_path, &rem_path, cmpl_state);

  if (!cmpl_state_okay (cmpl_state))
    {
      /* Something went wrong. */
      gtk_file_selection_abort (fs);
      return;
    }

  g_assert (cmpl_state->reference_dir);

  gtk_clist_freeze (GTK_CLIST (fs->dir_list));
  gtk_clist_clear (GTK_CLIST (fs->dir_list));
  gtk_clist_freeze (GTK_CLIST (fs->file_list));
  gtk_clist_clear (GTK_CLIST (fs->file_list));

  /* Set the dir_list to include ./ and ../ */
  text[1] = NULL;
  text[0] = "./";
  row = gtk_clist_append (GTK_CLIST (fs->dir_list), text);

  text[0] = "../";
  row = gtk_clist_append (GTK_CLIST (fs->dir_list), text);

  /*reset the max widths of the lists*/
  dir_list_width = gdk_string_width(fs->dir_list->style->font,"../");
  gtk_clist_set_column_width(GTK_CLIST(fs->dir_list),0,dir_list_width);
  file_list_width = 1;
  gtk_clist_set_column_width(GTK_CLIST(fs->file_list),0,file_list_width);

  while (poss)
    {
      if (cmpl_is_a_completion (poss))
        {
          possible_count += 1;

          filename = cmpl_this_completion (poss);

	  text[0] = filename;
	  
          if (cmpl_is_directory (poss))
            {
              if (strcmp (filename, "./") != 0 &&
                  strcmp (filename, "../") != 0)
		{
		  int width = gdk_string_width(fs->dir_list->style->font,
					       filename);
		  row = gtk_clist_append (GTK_CLIST (fs->dir_list), text);
		  if(width > dir_list_width)
		    {
		      dir_list_width = width;
		      gtk_clist_set_column_width(GTK_CLIST(fs->dir_list),0,
						 width);
		    }
 		}
	    }
          else
	    {
	      int width = gdk_string_width(fs->file_list->style->font,
				           filename);
	      row = gtk_clist_append (GTK_CLIST (fs->file_list), text);
	      if(width > file_list_width)
	        {
	          file_list_width = width;
	          gtk_clist_set_column_width(GTK_CLIST(fs->file_list),0,
					     width);
	        }
            }
	}

      poss = cmpl_next_completion (cmpl_state);
    }

  gtk_clist_thaw (GTK_CLIST (fs->dir_list));
  gtk_clist_thaw (GTK_CLIST (fs->file_list));

  /* File lists are set. */

  g_assert (cmpl_state->reference_dir);

  if (try_complete)
    {

      /* User is trying to complete filenames, so advance the user's input
       * string to the updated_text, which is the common leading substring
       * of all possible completions, and if its a directory attempt
       * attempt completions in it. */

      if (cmpl_updated_text (cmpl_state)[0])
        {

          if (cmpl_updated_dir (cmpl_state))
            {
	      gchar* dir_name = g_strdup (cmpl_updated_text (cmpl_state));

              did_recurse = TRUE;

              gtk_file_selection_populate (fs, dir_name, TRUE);

              g_free (dir_name);
            }
          else
            {
	      if (fs->selection_entry)
		      gtk_entry_set_text (GTK_ENTRY (fs->selection_entry),
					  cmpl_updated_text (cmpl_state));
            }
        }
      else
        {
          selection_index = cmpl_last_valid_char (cmpl_state) -
                            (strlen (rel_path) - strlen (rem_path));
	  if (fs->selection_entry)
	    gtk_entry_set_text (GTK_ENTRY (fs->selection_entry), rem_path);
        }
    }
  else
    {
      gboolean last_entry_was_dir_selection;

      last_entry_was_dir_selection = (gboolean) gtk_object_get_data (GTK_OBJECT (fs->selection_entry),
								     "last_entry_was_dir_selection");

      if (fs->selection_entry && last_entry_was_dir_selection)
	gtk_entry_set_text (GTK_ENTRY (fs->selection_entry), "");
    }

  if (!did_recurse)
    {
      if (fs->selection_entry)
	gtk_entry_set_position (GTK_ENTRY (fs->selection_entry), selection_index);

      if (fs->selection_entry)
	{
	  sel_text = g_strconcat (_("Selection: "),
				  cmpl_reference_position (cmpl_state),
				  NULL);

	  gtk_label_set_text (GTK_LABEL (fs->selection_text), sel_text);
	  g_free (sel_text);
	}

      if (fs->history_pulldown) 
	{
	  gtk_file_selection_update_history_menu (fs, cmpl_reference_position (cmpl_state));
	}

    }
}

static void
gtk_file_selection_abort (GtkFileSelection *fs)
{
  gchar err_buf[256];

  sprintf (err_buf, _("Directory unreadable: %s"), cmpl_strerror (cmpl_errno));

  /*  BEEP gdk_beep();  */

  if (fs->selection_entry)
    gtk_label_set_text (GTK_LABEL (fs->selection_text), err_buf);
}

/**********************************************************************/
/*			  External Interface                          */
/**********************************************************************/

/* The four completion state selectors
 */
static gchar*
cmpl_updated_text (CompletionState* cmpl_state)
{
  return cmpl_state->updated_text;
}

static gint
cmpl_updated_dir (CompletionState* cmpl_state)
{
  return cmpl_state->re_complete;
}

static gchar*
cmpl_reference_position (CompletionState* cmpl_state)
{
  return cmpl_state->reference_dir->fullname;
}

static gint
cmpl_last_valid_char (CompletionState* cmpl_state)
{
  return cmpl_state->last_valid_char;
}

static gchar*
cmpl_completion_fullname (gchar* text, CompletionState* cmpl_state)
{
  static char nothing[2] = "";

  if (!cmpl_state_okay (cmpl_state))
    {
      return nothing;
    }
  else if (text[0] == '/')
    {
      strcpy (cmpl_state->updated_text, text);
    }
  else if (text[0] == '~')
    {
      CompletionDir* dir;
      char* slash;

      dir = open_user_dir (text, cmpl_state);

      if (!dir)
	{
	  /* spencer says just return ~something, so
	   * for now just do it. */
	  strcpy (cmpl_state->updated_text, text);
	}
      else
	{

	  strcpy (cmpl_state->updated_text, dir->fullname);

	  slash = strchr (text, '/');

	  if (slash)
	    strcat (cmpl_state->updated_text, slash);
	}
    }
  else
    {
      strcpy (cmpl_state->updated_text, cmpl_state->reference_dir->fullname);
      if (strcmp (cmpl_state->reference_dir->fullname, "/") != 0)
	strcat (cmpl_state->updated_text, "/");
      strcat (cmpl_state->updated_text, text);
    }

  return cmpl_state->updated_text;
}

/* The three completion selectors
 */
static gchar*
cmpl_this_completion (PossibleCompletion* pc)
{
  return pc->text;
}

static gint
cmpl_is_directory (PossibleCompletion* pc)
{
  return pc->is_directory;
}

static gint
cmpl_is_a_completion (PossibleCompletion* pc)
{
  return pc->is_a_completion;
}

/**********************************************************************/
/*	                 Construction, deletion                       */
/**********************************************************************/

static CompletionState*
cmpl_init_state (void)
{
  gchar getcwd_buf[2*MAXPATHLEN];
  CompletionState *new_state;

  new_state = g_new (CompletionState, 1);

  /* We don't use getcwd() on SUNOS, because, it does a popen("pwd")
   * and, if that wasn't bad enough, hangs in doing so.
   */
#if defined(sun) && !defined(__SVR4)
  if (!getwd (getcwd_buf))
#else    
  if (!getcwd (getcwd_buf, MAXPATHLEN))
#endif    
    {
      /* Oh joy, we can't get the current directory. Um..., we should have
       * a root directory, right? Right? (Probably not portable to non-Unix)
       */
      strcpy (getcwd_buf, "/");
    }

tryagain:

  new_state->reference_dir = NULL;
  new_state->completion_dir = NULL;
  new_state->active_completion_dir = NULL;
  new_state->directory_storage = NULL;
  new_state->directory_sent_storage = NULL;
  new_state->last_valid_char = 0;
  new_state->updated_text = g_new (gchar, MAXPATHLEN);
  new_state->updated_text_alloc = MAXPATHLEN;
  new_state->the_completion.text = g_new (gchar, MAXPATHLEN);
  new_state->the_completion.text_alloc = MAXPATHLEN;
  new_state->user_dir_name_buffer = NULL;
  new_state->user_directories = NULL;

  new_state->reference_dir =  open_dir (getcwd_buf, new_state);

  if (!new_state->reference_dir)
    {
      /* Directories changing from underneath us, grumble */
      strcpy (getcwd_buf, "/");
      goto tryagain;
    }

  return new_state;
}

static void
cmpl_free_dir_list(GList* dp0)
{
  GList *dp = dp0;

  while (dp) {
    free_dir (dp->data);
    dp = dp->next;
  }

  g_list_free(dp0);
}

static void
cmpl_free_dir_sent_list(GList* dp0)
{
  GList *dp = dp0;

  while (dp) {
    free_dir_sent (dp->data);
    dp = dp->next;
  }

  g_list_free(dp0);
}

static void
cmpl_free_state (CompletionState* cmpl_state)
{
  cmpl_free_dir_list (cmpl_state->directory_storage);
  cmpl_free_dir_sent_list (cmpl_state->directory_sent_storage);

  if (cmpl_state->user_dir_name_buffer)
    g_free (cmpl_state->user_dir_name_buffer);
  if (cmpl_state->user_directories)
    g_free (cmpl_state->user_directories);
  if (cmpl_state->the_completion.text)
    g_free (cmpl_state->the_completion.text);
  if (cmpl_state->updated_text)
    g_free (cmpl_state->updated_text);

  g_free (cmpl_state);
}

static void
free_dir(CompletionDir* dir)
{
  g_free(dir->fullname);
  g_free(dir);
}

static void
free_dir_sent(CompletionDirSent* sent)
{
  g_free(sent->name_buffer);
  g_free(sent->entries);
  g_free(sent);
}

static void
prune_memory_usage(CompletionState *cmpl_state)
{
  GList* cdsl = cmpl_state->directory_sent_storage;
  GList* cdl = cmpl_state->directory_storage;
  GList* cdl0 = cdl;
  gint len = 0;

  for(; cdsl && len < CMPL_DIRECTORY_CACHE_SIZE; len += 1)
    cdsl = cdsl->next;

  if (cdsl) {
    cmpl_free_dir_sent_list(cdsl->next);
    cdsl->next = NULL;
  }

  cmpl_state->directory_storage = NULL;
  while (cdl) {
    if (cdl->data == cmpl_state->reference_dir)
      cmpl_state->directory_storage = g_list_prepend(NULL, cdl->data);
    else
      free_dir (cdl->data);
    cdl = cdl->next;
  }

  g_list_free(cdl0);
}

/**********************************************************************/
/*                        The main entrances.                         */
/**********************************************************************/

static PossibleCompletion*
cmpl_completion_matches (gchar* text_to_complete,
			 gchar** remaining_text,
			 CompletionState* cmpl_state)
{
  gchar* first_slash;
  PossibleCompletion *poss;

  prune_memory_usage(cmpl_state);

  g_assert (text_to_complete != NULL);

  cmpl_state->user_completion_index = -1;
  cmpl_state->last_completion_text = text_to_complete;
  cmpl_state->the_completion.text[0] = 0;
  cmpl_state->last_valid_char = 0;
  cmpl_state->updated_text_len = -1;
  cmpl_state->updated_text[0] = 0;
  cmpl_state->re_complete = FALSE;

  first_slash = strchr (text_to_complete, '/');

  if (text_to_complete[0] == '~' && !first_slash)
    {
      /* Text starts with ~ and there is no slash, show all the
       * home directory completions.
       */
      poss = attempt_homedir_completion (text_to_complete, cmpl_state);

      update_cmpl(poss, cmpl_state);

      return poss;
    }

  cmpl_state->reference_dir =
    open_ref_dir (text_to_complete, remaining_text, cmpl_state);

  if(!cmpl_state->reference_dir)
    return NULL;

  cmpl_state->completion_dir =
    find_completion_dir (*remaining_text, remaining_text, cmpl_state);

  cmpl_state->last_valid_char = *remaining_text - text_to_complete;

  if(!cmpl_state->completion_dir)
    return NULL;

  cmpl_state->completion_dir->cmpl_index = -1;
  cmpl_state->completion_dir->cmpl_parent = NULL;
  cmpl_state->completion_dir->cmpl_text = *remaining_text;

  cmpl_state->active_completion_dir = cmpl_state->completion_dir;

  cmpl_state->reference_dir = cmpl_state->completion_dir;

  poss = attempt_file_completion(cmpl_state);

  update_cmpl(poss, cmpl_state);

  return poss;
}

static PossibleCompletion*
cmpl_next_completion (CompletionState* cmpl_state)
{
  PossibleCompletion* poss = NULL;

  cmpl_state->the_completion.text[0] = 0;

  if(cmpl_state->user_completion_index >= 0)
    poss = attempt_homedir_completion(cmpl_state->last_completion_text, cmpl_state);
  else
    poss = attempt_file_completion(cmpl_state);

  update_cmpl(poss, cmpl_state);

  return poss;
}

/**********************************************************************/
/*			 Directory Operations                         */
/**********************************************************************/

/* Open the directory where completion will begin from, if possible. */
static CompletionDir*
open_ref_dir(gchar* text_to_complete,
	     gchar** remaining_text,
	     CompletionState* cmpl_state)
{
  gchar* first_slash;
  CompletionDir *new_dir;

  first_slash = strchr(text_to_complete, '/');

  if (text_to_complete[0] == '~')
    {
      new_dir = open_user_dir(text_to_complete, cmpl_state);

      if(new_dir)
	{
	  if(first_slash)
	    *remaining_text = first_slash + 1;
	  else
	    *remaining_text = text_to_complete + strlen(text_to_complete);
	}
      else
	{
	  return NULL;
	}
    }
  else if (text_to_complete[0] == '/' || !cmpl_state->reference_dir)
    {
      gchar *tmp = g_strdup(text_to_complete);
      gchar *p;

      p = tmp;
      while (*p && *p != '*' && *p != '?')
	p++;

      *p = '\0';
      p = strrchr(tmp, '/');
      if (p)
	{
	  if (p == tmp)
	    p++;
      
	  *p = '\0';

	  new_dir = open_dir(tmp, cmpl_state);

	  if(new_dir)
	    *remaining_text = text_to_complete + 
	      ((p == tmp + 1) ? (p - tmp) : (p + 1 - tmp));
	}
      else
	{
	  /* If no possible candidates, use the cwd */
	  gchar *curdir = g_get_current_dir ();
	  
	  new_dir = open_dir(curdir, cmpl_state);

	  if (new_dir)
	    *remaining_text = text_to_complete;

	  g_free (curdir);
	}

      g_free (tmp);
    }
  else
    {
      *remaining_text = text_to_complete;

      new_dir = open_dir(cmpl_state->reference_dir->fullname, cmpl_state);
    }

  if(new_dir)
    {
      new_dir->cmpl_index = -1;
      new_dir->cmpl_parent = NULL;
    }

  return new_dir;
}

/* open a directory by user name */
static CompletionDir*
open_user_dir(gchar* text_to_complete,
	      CompletionState *cmpl_state)
{
  gchar *first_slash;
  gint cmp_len;

  g_assert(text_to_complete && text_to_complete[0] == '~');

  first_slash = strchr(text_to_complete, '/');

  if (first_slash)
    cmp_len = first_slash - text_to_complete - 1;
  else
    cmp_len = strlen(text_to_complete + 1);

  if(!cmp_len)
    {
      /* ~/ */
      gchar *homedir = g_get_home_dir ();

      if (homedir)
	return open_dir(homedir, cmpl_state);
      else
	return NULL;
    }
  else
    {
      /* ~user/ */
      char* copy = g_new(char, cmp_len + 1);
      struct passwd *pwd;
      strncpy(copy, text_to_complete + 1, cmp_len);
      copy[cmp_len] = 0;
      pwd = getpwnam(copy);
      g_free(copy);
      if (!pwd)
	{
	  cmpl_errno = errno;
	  return NULL;
	}

      return open_dir(pwd->pw_dir, cmpl_state);
    }
}

/* open a directory relative the the current relative directory */
static CompletionDir*
open_relative_dir(gchar* dir_name,
		  CompletionDir* dir,
		  CompletionState *cmpl_state)
{
  gchar path_buf[2*MAXPATHLEN];

  if(dir->fullname_len + strlen(dir_name) + 2 >= MAXPATHLEN)
    {
      cmpl_errno = CMPL_ERRNO_TOO_LONG;
      return NULL;
    }

  strcpy(path_buf, dir->fullname);

  if(dir->fullname_len > 1)
    {
      path_buf[dir->fullname_len] = '/';
      strcpy(path_buf + dir->fullname_len + 1, dir_name);
    }
  else
    {
      strcpy(path_buf + dir->fullname_len, dir_name);
    }

  return open_dir(path_buf, cmpl_state);
}

/* after the cache lookup fails, really open a new directory */
static CompletionDirSent*
open_new_dir(gchar* dir_name, struct stat* sbuf, gboolean stat_subdirs)
{
  CompletionDirSent* sent;
  DIR* directory;
  gchar *buffer_ptr;
  struct dirent *dirent_ptr;
  gint buffer_size = 0;
  gint entry_count = 0;
  gint i;
  struct stat ent_sbuf;
  char path_buf[MAXPATHLEN*2];
  gint path_buf_len;

  sent = g_new(CompletionDirSent, 1);
  sent->mtime = sbuf->st_mtime;
  sent->inode = sbuf->st_ino;
  sent->device = sbuf->st_dev;

  path_buf_len = strlen(dir_name);

  if (path_buf_len > MAXPATHLEN)
    {
      cmpl_errno = CMPL_ERRNO_TOO_LONG;
      return NULL;
    }

  strcpy(path_buf, dir_name);

  directory = opendir(dir_name);

  if(!directory)
    {
      cmpl_errno = errno;
      return NULL;
    }

  while((dirent_ptr = readdir(directory)) != NULL)
    {
      int entry_len = strlen(dirent_ptr->d_name);
      buffer_size += entry_len + 1;
      entry_count += 1;

      if(path_buf_len + entry_len + 2 >= MAXPATHLEN)
	{
	  cmpl_errno = CMPL_ERRNO_TOO_LONG;
 	  closedir(directory);
	  return NULL;
	}
    }

  sent->name_buffer = g_new(gchar, buffer_size);
  sent->entries = g_new(CompletionDirEntry, entry_count);
  sent->entry_count = entry_count;

  buffer_ptr = sent->name_buffer;

  rewinddir(directory);

  for(i = 0; i < entry_count; i += 1)
    {
      dirent_ptr = readdir(directory);

      if(!dirent_ptr)
	{
	  cmpl_errno = errno;
	  closedir(directory);
	  return NULL;
	}

      strcpy(buffer_ptr, dirent_ptr->d_name);
      sent->entries[i].entry_name = buffer_ptr;
      buffer_ptr += strlen(dirent_ptr->d_name);
      *buffer_ptr = 0;
      buffer_ptr += 1;

      path_buf[path_buf_len] = '/';
      strcpy(path_buf + path_buf_len + 1, dirent_ptr->d_name);

      if (stat_subdirs)
	{
	  if(stat(path_buf, &ent_sbuf) >= 0 && S_ISDIR(ent_sbuf.st_mode))
	    sent->entries[i].is_dir = 1;
	  else
	    /* stat may fail, and we don't mind, since it could be a
	     * dangling symlink. */
	    sent->entries[i].is_dir = 0;
	}
      else
	sent->entries[i].is_dir = 1;
    }

  qsort(sent->entries, sent->entry_count, sizeof(CompletionDirEntry), compare_cmpl_dir);

  closedir(directory);

  return sent;
}

static gboolean
check_dir(gchar *dir_name, struct stat *result, gboolean *stat_subdirs)
{
  /* A list of directories that we know only contain other directories.
   * Trying to stat every file in these directories would be very
   * expensive.
   */

  static struct {
    gchar *name;
    gboolean present;
    struct stat statbuf;
  } no_stat_dirs[] = {
    { "/afs", FALSE, { 0 } },
    { "/net", FALSE, { 0 } }
  };

  static const gint n_no_stat_dirs = sizeof(no_stat_dirs) / sizeof(no_stat_dirs[0]);
  static gboolean initialized = FALSE;

  gint i;

  if (!initialized)
    {
      initialized = TRUE;
      for (i = 0; i < n_no_stat_dirs; i++)
	{
	  if (stat (no_stat_dirs[i].name, &no_stat_dirs[i].statbuf) == 0)
	    no_stat_dirs[i].present = TRUE;
	}
    }

  if(stat(dir_name, result) < 0)
    {
      cmpl_errno = errno;
      return FALSE;
    }

  *stat_subdirs = TRUE;
  for (i=0; i<n_no_stat_dirs; i++)
    {
      if (no_stat_dirs[i].present &&
	  (no_stat_dirs[i].statbuf.st_dev == result->st_dev) &&
	  (no_stat_dirs[i].statbuf.st_ino == result->st_ino))
	{
	  *stat_subdirs = FALSE;
	  break;
	}
    }

  return TRUE;
}

/* open a directory by absolute pathname */
static CompletionDir*
open_dir(gchar* dir_name, CompletionState* cmpl_state)
{
  struct stat sbuf;
  gboolean stat_subdirs;
  CompletionDirSent *sent;
  GList* cdsl;

  if (!check_dir (dir_name, &sbuf, &stat_subdirs))
    return NULL;

  cdsl = cmpl_state->directory_sent_storage;

  while (cdsl)
    {
      sent = cdsl->data;

      if(sent->inode == sbuf.st_ino &&
	 sent->mtime == sbuf.st_mtime &&
	 sent->device == sbuf.st_dev)
	return attach_dir(sent, dir_name, cmpl_state);

      cdsl = cdsl->next;
    }

  sent = open_new_dir(dir_name, &sbuf, stat_subdirs);

  if (sent) {
    cmpl_state->directory_sent_storage =
      g_list_prepend(cmpl_state->directory_sent_storage, sent);

    return attach_dir(sent, dir_name, cmpl_state);
  }

  return NULL;
}

static CompletionDir*
attach_dir(CompletionDirSent* sent, gchar* dir_name, CompletionState *cmpl_state)
{
  CompletionDir* new_dir;

  new_dir = g_new(CompletionDir, 1);

  cmpl_state->directory_storage =
    g_list_prepend(cmpl_state->directory_storage, new_dir);

  new_dir->sent = sent;
  new_dir->fullname = g_strdup(dir_name);
  new_dir->fullname_len = strlen(dir_name);

  return new_dir;
}

static gint
correct_dir_fullname(CompletionDir* cmpl_dir)
{
  gint length = strlen(cmpl_dir->fullname);
  struct stat sbuf;

  if (strcmp(cmpl_dir->fullname + length - 2, "/.") == 0)
    {
      if (length == 2) 
	{
	  strcpy(cmpl_dir->fullname, "/");
	  cmpl_dir->fullname_len = 1;
	  return TRUE;
	} else {
	  cmpl_dir->fullname[length - 2] = 0;
	}
    }
  else if (strcmp(cmpl_dir->fullname + length - 3, "/./") == 0)
    cmpl_dir->fullname[length - 2] = 0;
  else if (strcmp(cmpl_dir->fullname + length - 3, "/..") == 0)
    {
      if(length == 3)
	{
	  strcpy(cmpl_dir->fullname, "/");
	  cmpl_dir->fullname_len = 1;
	  return TRUE;
	}

      if(stat(cmpl_dir->fullname, &sbuf) < 0)
	{
	  cmpl_errno = errno;
	  return FALSE;
	}

      cmpl_dir->fullname[length - 2] = 0;

      if(!correct_parent(cmpl_dir, &sbuf))
	return FALSE;
    }
  else if (strcmp(cmpl_dir->fullname + length - 4, "/../") == 0)
    {
      if(length == 4)
	{
	  strcpy(cmpl_dir->fullname, "/");
	  cmpl_dir->fullname_len = 1;
	  return TRUE;
	}

      if(stat(cmpl_dir->fullname, &sbuf) < 0)
	{
	  cmpl_errno = errno;
	  return FALSE;
	}

      cmpl_dir->fullname[length - 3] = 0;

      if(!correct_parent(cmpl_dir, &sbuf))
	return FALSE;
    }

  cmpl_dir->fullname_len = strlen(cmpl_dir->fullname);

  return TRUE;
}

static gint
correct_parent(CompletionDir* cmpl_dir, struct stat *sbuf)
{
  struct stat parbuf;
  gchar *last_slash;
  gchar *new_name;
  gchar c = 0;

  last_slash = strrchr(cmpl_dir->fullname, '/');

  g_assert(last_slash);

  if(last_slash != cmpl_dir->fullname)
    { /* last_slash[0] = 0; */ }
  else
    {
      c = last_slash[1];
      last_slash[1] = 0;
    }

  if (stat(cmpl_dir->fullname, &parbuf) < 0)
    {
      cmpl_errno = errno;
      return FALSE;
    }

  if (parbuf.st_ino == sbuf->st_ino && parbuf.st_dev == sbuf->st_dev)
    /* it wasn't a link */
    return TRUE;

  if(c)
    last_slash[1] = c;
  /* else
    last_slash[0] = '/'; */

  /* it was a link, have to figure it out the hard way */

  new_name = find_parent_dir_fullname(cmpl_dir->fullname);

  if (!new_name)
    return FALSE;

  g_free(cmpl_dir->fullname);

  cmpl_dir->fullname = new_name;

  return TRUE;
}

static gchar*
find_parent_dir_fullname(gchar* dirname)
{
  gchar buffer[MAXPATHLEN];
  gchar buffer2[MAXPATHLEN];

#if defined(sun) && !defined(__SVR4)
  if(!getwd(buffer))
#else
  if(!getcwd(buffer, MAXPATHLEN))
#endif    
    {
      cmpl_errno = errno;
      return NULL;
    }

  if(chdir(dirname) != 0 || chdir("..") != 0)
    {
      cmpl_errno = errno;
      return NULL;
    }

#if defined(sun) && !defined(__SVR4)
  if(!getwd(buffer2))
#else
  if(!getcwd(buffer2, MAXPATHLEN))
#endif
    {
      chdir(buffer);
      cmpl_errno = errno;

      return NULL;
    }

  if(chdir(buffer) != 0)
    {
      cmpl_errno = errno;
      return NULL;
    }

  return g_strdup(buffer2);
}

/**********************************************************************/
/*                        Completion Operations                       */
/**********************************************************************/

static PossibleCompletion*
attempt_homedir_completion(gchar* text_to_complete,
			   CompletionState *cmpl_state)
{
  gint index, length;

  if (!cmpl_state->user_dir_name_buffer &&
      !get_pwdb(cmpl_state))
    return NULL;
  length = strlen(text_to_complete) - 1;

  cmpl_state->user_completion_index += 1;

  while(cmpl_state->user_completion_index < cmpl_state->user_directories_len)
    {
      index = first_diff_index(text_to_complete + 1,
			       cmpl_state->user_directories
			       [cmpl_state->user_completion_index].login);

      switch(index)
	{
	case PATTERN_MATCH:
	  break;
	default:
	  if(cmpl_state->last_valid_char < (index + 1))
	    cmpl_state->last_valid_char = index + 1;
	  cmpl_state->user_completion_index += 1;
	  continue;
	}

      cmpl_state->the_completion.is_a_completion = 1;
      cmpl_state->the_completion.is_directory = 1;

      append_completion_text("~", cmpl_state);

      append_completion_text(cmpl_state->
			      user_directories[cmpl_state->user_completion_index].login,
			     cmpl_state);

      return append_completion_text("/", cmpl_state);
    }

  if(text_to_complete[1] ||
     cmpl_state->user_completion_index > cmpl_state->user_directories_len)
    {
      cmpl_state->user_completion_index = -1;
      return NULL;
    }
  else
    {
      cmpl_state->user_completion_index += 1;
      cmpl_state->the_completion.is_a_completion = 1;
      cmpl_state->the_completion.is_directory = 1;

      return append_completion_text("~/", cmpl_state);
    }
}

/* returns the index (>= 0) of the first differing character,
 * PATTERN_MATCH if the completion matches */
static gint
first_diff_index(gchar* pat, gchar* text)
{
  gint diff = 0;

  while(*pat && *text && *text == *pat)
    {
      pat += 1;
      text += 1;
      diff += 1;
    }

  if(*pat)
    return diff;

  return PATTERN_MATCH;
}

static PossibleCompletion*
append_completion_text(gchar* text, CompletionState* cmpl_state)
{
  gint len, i = 1;

  if(!cmpl_state->the_completion.text)
    return NULL;

  len = strlen(text) + strlen(cmpl_state->the_completion.text) + 1;

  if(cmpl_state->the_completion.text_alloc > len)
    {
      strcat(cmpl_state->the_completion.text, text);
      return &cmpl_state->the_completion;
    }

  while(i < len) { i <<= 1; }

  cmpl_state->the_completion.text_alloc = i;

  cmpl_state->the_completion.text = (gchar*)g_realloc(cmpl_state->the_completion.text, i);

  if(!cmpl_state->the_completion.text)
    return NULL;
  else
    {
      strcat(cmpl_state->the_completion.text, text);
      return &cmpl_state->the_completion;
    }
}

static CompletionDir*
find_completion_dir(gchar* text_to_complete,
		    gchar** remaining_text,
		    CompletionState* cmpl_state)
{
  gchar* first_slash = strchr(text_to_complete, '/');
  CompletionDir* dir = cmpl_state->reference_dir;
  CompletionDir* next;
  *remaining_text = text_to_complete;

  while(first_slash)
    {
      gint len = first_slash - *remaining_text;
      gint found = 0;
      gchar *found_name = NULL;         /* Quiet gcc */
      gint i;
      gchar* pat_buf = g_new (gchar, len + 1);

      strncpy(pat_buf, *remaining_text, len);
      pat_buf[len] = 0;

      for(i = 0; i < dir->sent->entry_count; i += 1)
	{
	  if(dir->sent->entries[i].is_dir &&
	     fnmatch(pat_buf, dir->sent->entries[i].entry_name,
		     FNMATCH_FLAGS)!= FNM_NOMATCH)
	    {
	      if(found)
		{
		  g_free (pat_buf);
		  return dir;
		}
	      else
		{
		  found = 1;
		  found_name = dir->sent->entries[i].entry_name;
		}
	    }
	}

      if (!found)
	{
	  /* Perhaps we are trying to open an automount directory */
	  found_name = pat_buf;
	}

      next = open_relative_dir(found_name, dir, cmpl_state);
      
      if(!next)
	{
	  g_free (pat_buf);
	  return NULL;
	}
      
      next->cmpl_parent = dir;
      
      dir = next;
      
      if(!correct_dir_fullname(dir))
	{
	  g_free(pat_buf);
	  return NULL;
	}
      
      *remaining_text = first_slash + 1;
      first_slash = strchr(*remaining_text, '/');

      g_free (pat_buf);
    }

  return dir;
}

static void
update_cmpl(PossibleCompletion* poss, CompletionState* cmpl_state)
{
  gint cmpl_len;

  if(!poss || !cmpl_is_a_completion(poss))
    return;

  cmpl_len = strlen(cmpl_this_completion(poss));

  if(cmpl_state->updated_text_alloc < cmpl_len + 1)
    {
      cmpl_state->updated_text =
	(gchar*)g_realloc(cmpl_state->updated_text,
			  cmpl_state->updated_text_alloc);
      cmpl_state->updated_text_alloc = 2*cmpl_len;
    }

  if(cmpl_state->updated_text_len < 0)
    {
      strcpy(cmpl_state->updated_text, cmpl_this_completion(poss));
      cmpl_state->updated_text_len = cmpl_len;
      cmpl_state->re_complete = cmpl_is_directory(poss);
    }
  else if(cmpl_state->updated_text_len == 0)
    {
      cmpl_state->re_complete = FALSE;
    }
  else
    {
      gint first_diff =
	first_diff_index(cmpl_state->updated_text,
			 cmpl_this_completion(poss));

      cmpl_state->re_complete = FALSE;

      if(first_diff == PATTERN_MATCH)
	return;

      if(first_diff > cmpl_state->updated_text_len)
	strcpy(cmpl_state->updated_text, cmpl_this_completion(poss));

      cmpl_state->updated_text_len = first_diff;
      cmpl_state->updated_text[first_diff] = 0;
    }
}

static PossibleCompletion*
attempt_file_completion(CompletionState *cmpl_state)
{
  gchar *pat_buf, *first_slash;
  CompletionDir *dir = cmpl_state->active_completion_dir;

  dir->cmpl_index += 1;

  if(dir->cmpl_index == dir->sent->entry_count)
    {
      if(dir->cmpl_parent == NULL)
	{
	  cmpl_state->active_completion_dir = NULL;

	  return NULL;
	}
      else
	{
	  cmpl_state->active_completion_dir = dir->cmpl_parent;

	  return attempt_file_completion(cmpl_state);
	}
    }

  g_assert(dir->cmpl_text);

  first_slash = strchr(dir->cmpl_text, '/');

  if(first_slash)
    {
      gint len = first_slash - dir->cmpl_text;

      pat_buf = g_new (gchar, len + 1);
      strncpy(pat_buf, dir->cmpl_text, len);
      pat_buf[len] = 0;
    }
  else
    {
      gint len = strlen(dir->cmpl_text);

      pat_buf = g_new (gchar, len + 2);
      strcpy(pat_buf, dir->cmpl_text);
      strcpy(pat_buf + len, "*");
    }

  if(first_slash)
    {
      if(dir->sent->entries[dir->cmpl_index].is_dir)
	{
	  if(fnmatch(pat_buf, dir->sent->entries[dir->cmpl_index].entry_name,
		     FNMATCH_FLAGS) != FNM_NOMATCH)
	    {
	      CompletionDir* new_dir;

	      new_dir = open_relative_dir(dir->sent->entries[dir->cmpl_index].entry_name,
					  dir, cmpl_state);

	      if(!new_dir)
		{
		  g_free (pat_buf);
		  return NULL;
		}

	      new_dir->cmpl_parent = dir;

	      new_dir->cmpl_index = -1;
	      new_dir->cmpl_text = first_slash + 1;

	      cmpl_state->active_completion_dir = new_dir;

	      g_free (pat_buf);
	      return attempt_file_completion(cmpl_state);
	    }
	  else
	    {
	      g_free (pat_buf);
	      return attempt_file_completion(cmpl_state);
	    }
	}
      else
	{
	  g_free (pat_buf);
	  return attempt_file_completion(cmpl_state);
	}
    }
  else
    {
      if(dir->cmpl_parent != NULL)
	{
	  append_completion_text(dir->fullname +
				 strlen(cmpl_state->completion_dir->fullname) + 1,
				 cmpl_state);
	  append_completion_text("/", cmpl_state);
	}

      append_completion_text(dir->sent->entries[dir->cmpl_index].entry_name, cmpl_state);

      cmpl_state->the_completion.is_a_completion =
	(fnmatch(pat_buf, dir->sent->entries[dir->cmpl_index].entry_name,
		 FNMATCH_FLAGS) != FNM_NOMATCH);

      cmpl_state->the_completion.is_directory = dir->sent->entries[dir->cmpl_index].is_dir;
      if(dir->sent->entries[dir->cmpl_index].is_dir)
	append_completion_text("/", cmpl_state);

      g_free (pat_buf);
      return &cmpl_state->the_completion;
    }
}


static gint
get_pwdb(CompletionState* cmpl_state)
{
  struct passwd *pwd_ptr;
  gchar* buf_ptr;
  gint len = 0, i, count = 0;

  if(cmpl_state->user_dir_name_buffer)
    return TRUE;
  setpwent ();

  while ((pwd_ptr = getpwent()) != NULL)
    {
      len += strlen(pwd_ptr->pw_name);
      len += strlen(pwd_ptr->pw_dir);
      len += 2;
      count += 1;
    }

  setpwent ();

  cmpl_state->user_dir_name_buffer = g_new(gchar, len);
  cmpl_state->user_directories = g_new(CompletionUserDir, count);
  cmpl_state->user_directories_len = count;

  buf_ptr = cmpl_state->user_dir_name_buffer;

  for(i = 0; i < count; i += 1)
    {
      pwd_ptr = getpwent();
      if(!pwd_ptr)
	{
	  cmpl_errno = errno;
	  goto error;
	}

      strcpy(buf_ptr, pwd_ptr->pw_name);
      cmpl_state->user_directories[i].login = buf_ptr;
      buf_ptr += strlen(buf_ptr);
      buf_ptr += 1;
      strcpy(buf_ptr, pwd_ptr->pw_dir);
      cmpl_state->user_directories[i].homedir = buf_ptr;
      buf_ptr += strlen(buf_ptr);
      buf_ptr += 1;
    }

  qsort(cmpl_state->user_directories,
	cmpl_state->user_directories_len,
	sizeof(CompletionUserDir),
	compare_user_dir);

  endpwent();

  return TRUE;

error:

  if(cmpl_state->user_dir_name_buffer)
    g_free(cmpl_state->user_dir_name_buffer);
  if(cmpl_state->user_directories)
    g_free(cmpl_state->user_directories);

  cmpl_state->user_dir_name_buffer = NULL;
  cmpl_state->user_directories = NULL;

  return FALSE;
}

static gint
compare_user_dir(const void* a, const void* b)
{
  return strcmp((((CompletionUserDir*)a))->login,
		(((CompletionUserDir*)b))->login);
}

static gint
compare_cmpl_dir(const void* a, const void* b)
{
  return strcmp((((CompletionDirEntry*)a))->entry_name,
		(((CompletionDirEntry*)b))->entry_name);
}

static gint
cmpl_state_okay(CompletionState* cmpl_state)
{
  return  cmpl_state && cmpl_state->reference_dir;
}

static gchar*
cmpl_strerror(gint err)
{
  if(err == CMPL_ERRNO_TOO_LONG)
    return "Name too long";
  else
    return g_strerror (err);
}
